import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.startup_routines import dependency_injection
dependency_injection()
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi

db = default_DB()

r_base_path = '/gpfs/exfel/theory_group/user/berberic/MTIP/reconstructions/20_8_2022/run_0/'
opt = db.load_settings('analysis',path = r_base_path + 'analysis.yaml')
base_path = '/gpfs/exfel/theory_group/user/berberic/MTIP/reciprocal_proj_data/'
path = base_path + '3d_covid_1p_10k_mO70_02_psd.h5'
data = db.load(path)

Bl = data['deg_2_invariant']
qs = data['data_radial_points']
options = {'plot_range':[1e8,1e17]}
out_path = '/gpfs/exfel/theory_group/user/berberic/MTIP/test/'

point_1=np.array((4,qs[3]))
point_2=np.array((70,0.34))
diff = point_2 - point_1
rot=np.array([[0,1],[-1,0]])
diff_rot = -(rot@diff)
grid = GridFactory.construct_grid('uniform',[np.arange(0,71),qs])
grid_2 = grid[:]-point_1
dot = np.sum(grid_2[:]*diff_rot[None,None,:],axis = -1)
mask = dot<0
mask = mask[:,:,None] + mask[:,None,:]
#Bl[mask] = 0

#print_bl = np.moveaxis(Bl,0,2)
#print_bl[:,mask.T]=0


def _save_first_invariants(bls,radial_points,base_path,options,name='diag_'):
    max_value = np.max(tuple(np.abs(bl).max() for bl in bls))
    vmin,vmax = options.get('plot_range',[max_value*1e-12,max_value])
    if [vmin,vmax] == [None,None]:
        vmin,vmax = [max_value*1e-12,max_value]
        
    grid = GridFactory.construct_grid('uniform',[radial_points,np.arange(bls.shape[2])])
    order_ids = [6,12,18,24,30]
    #order_ids = [20,40,60,80,127]
    
    shape = [2,len(order_ids)]
    layouts = []
    for i in range(shape[0]):
        layout_part = []
        orders = np.arange(10*i,10*i+len(order_ids)*2,2)
        for o in orders:
            layout = {'title':'$B_{'+'{}'.format(o)+'}$',
                      'x_label':'$q_1$',
                      'y_label':'$q_2$'
                      }
            layout_part.append(layout)
        layouts.append(layout_part)
                                                
    fig_bl_masks = heat2D_multi.get_fig([[np.abs(bls[n]) for n in order_ids],[np.abs(bls[10+n]) for n in order_ids] ],scale = 'log',layout = layouts,grid =grid[:],shape = shape,size = (30,10),vmin= vmin, vmax = vmax,cmap='plasma')
    bl_path = base_path +name+'Bl.matplotlib'
    db.save(bl_path,fig_bl_masks,dpi = 300)
    

#_save_first_invariants(print_bl[...,::2],qs,out_path,options,'diag_')
#_save_first_invariants(Bl,qs,out_path,options,'diag_2')



import xframe.library.mathLibrary as math
import xframe.library.pythonLibrary as python
import xframe.library.physicsLibrary as physics
import xframe.library.units
import xframe.library.gridLibrary as grid



import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_ft
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants
from xframe.library.mathLibrary import SampleShapeFunctions
from xframe.startup_routines import load_recipes
from xframe.control.Control import Controller
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_1p_00')[0]
controller = Controller(analysis_sketch)
from xframe.settings import analysis as opt
from xframe.database import analysis as db

opt.dimensions = 3
opt.grid.n_phis = 2*opt.grid.max_order+1
#opt.grid.n_angular_points = 100
pi_in_q = opt.fourier_transform.pi_in_q
if pi_in_q:
    name_postfix='N'+str(opt.grid.n_radial_points)+'mO'+str(opt.grid.max_order)+'nO'+str(len(opt.fourier_transform.pos_orders))+'_pi'
else:
    name_postfix='N'+str(opt.grid.n_radial_points)+'mO'+str(opt.grid.max_order)+'nO'+str(len(opt.fourier_transform.pos_orders))

#try:
#    weights = db.load('ft_weights',path_modifiers={'postfix':name_postfix,'type':opt.fourier_transform.type+'_'+str(opt.dimensions)+'D'})
#except FileNotFoundError as e:
#    weights = generate_weightDict_zernike(opt.grid.max_order,opt.grid.n_radial_points,dimensions=opt.dimensions)
mode = 'trapz'
weights = generate_weightDict(opt.grid.max_order,opt.grid.n_radial_points,dimensions=opt.dimensions,mode=mode)
    
data = db.load('reciprocal_proj_data',path_modifiers={'name':opt.name})
max_q= data['data_radial_points'].max()
max_r = np.pi*opt.grid.n_radial_points/max_q


ht_opt={
            'dimensions':opt.dimensions,
            **opt.grid,            
        }

harm_trf = HarmonicTransform('complex',ht_opt)
ft,ift =  generate_ft(max_r,weights,harm_trf,opt.dimensions,use_gpu=True,pi_in_q=pi_in_q,mode=mode)


grid_pair=get_grid({**opt.fourier_transform,**ht_opt,**harm_trf.grid_param,'max_q':max_q,'n_radial_points_from_data':256})
real_grid = grid_pair.realGrid

if opt.dimensions == 2:
    f = SampleShapeFunctions.get_polygon_function(600,6,coordSys = 'polar')
    density = f(real_grid[:]).astype(complex)
    density[:20,:19]=-1
if opt.dimensions ==3 :
    f = SampleShapeFunctions.get_tetrahedral_function(600)
    density = f(real_grid[:]).astype(complex)
    density[:20,:19,-10:]=-1
    
B1 = density_to_deg2_invariants(density,harm_trf,ft,opt.dimensions)
rd = ft(density)

d2 = ift(rd)
B2 = density_to_deg2_invariants(d2,harm_trf,ft,opt.dimensions)
diff = (density.real - d2.real)**2

if opt.dimensions == 3:
    db.save('fourier_test_{}_3D_2.vts'.format(mode),[density.real, d2.real , diff],grid = real_grid[:],dset_names=["original",'transformed', 'diff'],grid_type='spherical')
else:
    db.save('fourier_test_{}_2D.vts'.format(mode),[density.real, d2.real , diff],grid = real_grid[:],dset_names=["original",'transformed', 'diff'],grid_type='polar')
#    

qs=grid_pair.reciprocalGrid[:].__getitem__((slice(None),)+(0,)*opt.dimensions) # expression is just grid[:,0,0,0] in 3D case ans grid[:,0,0] in 2D
qqs=qs[None,:,None]*qs[None,None,:]
if opt.dimensions == 3:
    qqs *= qqs


diff_norm = np.sum((qqs*(B1-B2))**2,axis = (1,2))
norm = np.sum((qqs*B1)**2,axis = (1,2))
errors = np.full(len(norm),-1,dtype = float)
non_zero_mask = norm!=0.0
errors[non_zero_mask]=diff_norm[non_zero_mask]/norm[non_zero_mask]
print(errors)



orders = np.arange(71)
signs = [ (-1)**m for l in orders for m in range(1,l+1)]
pos_ids =  [ l*(l+1)+m for l in orders for m in range(1,l+1)]
neg_ids =  [ l*(l+1)+m for l in orders for m in range(-1,-l-1,-1)]
zero_ids = [ l*(l+1) for l in orders]
L = orders[-1]
pos_mask = np.zeros((L+1)**2,dtype = bool)
neg_mask = np.zeros((L+1)**2,dtype = bool)
pos_mask[pos_ids] = True
neg_mask[:] = ~pos_mask
neg_mask[zero_ids] = False
harm_rd = harm_trf.forward(rd)

herrors = [np.max(np.abs(harm_rd[l][19] - (-1)**l * (-1.0)**np.arange(-l,l+1)*harm_rd[l][19][::-1].conj())) for l in orders]
error = np.max(herrors)





import os
os.environ["MKL_NUM_THREADS"] = "1" 
os.environ["NUMEXPR_NUM_THREADS"] = "1" 
os.environ["OMP_NUM_THREADS"] = "1" 
os.environ['OPENBLAS_NUM_THREADS'] = "1"


import numpy as np
import scipy
import logging
import traceback
from xframe import log
os.chdir(os.path.expanduser('~/Programs/xframe'))
from xframe.startup_routines import load_recipes
from xframe.startup_routines import dependency_injection_no_soft
from xframe.library import mathLibrary
from xframe.externalLibraries.flt_plugin import LegendreTransform
mathLibrary.leg_trf = LegendreTransform
from xframe.externalLibraries.shtns_plugin import sh
mathLibrary.shtns = sh
from xframe.control.Control import Controller
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_to_deg2_invariant
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_remove_0_order
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import cross_correlation_mask
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_apply_precision_filter
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import ccd_to_deg2_invariant_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import bl_3d_pseudo_inverse_worker
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_cc_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import density_to_deg2_invariants_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_to_projection_matrices_3d
from xframe.plugins.MTIP.analysisLibrary.fxs_invariant_tools import deg2_invariant_eigenvalues
from xframe.plugins.MTIP.analysisLibrary import fxs_invariant_tools as i_tools
from xframe.plugins.MTIP.analysisLibrary.hankel_transforms import generate_weightDict_zernike_spherical
from xframe.plugins.MTIP.analysisLibrary.harmonic_transforms import HarmonicTransform
from xframe.plugins.MTIP.analysisLibrary.fourier_transforms import generate_zernike_spherical_ft
from xframe.plugins.MTIP.analysisLibrary.ft_grid_pairs import get_grid
from xframe.plugins.MTIP.analysisLibrary.fxs_Projections import ReciprocalProjection_old,ReciprocalProjection
from xframe import settings
from xframe.library import mathLibrary as mLib
from xframe import Multiprocessing
from xframe.library import physicsLibrary as pLib
from xframe.library.gridLibrary import GridFactory
from xframe.presenters.matplolibPresenter import heat2D_multi 
log=log.setup_custom_logger('root','INFO')

dependency_injection_no_soft()
analysis_sketch = load_recipes(ana_name='MTIP.reconstruct', ana_settings='3d_pentagon_100')[0]
db = analysis_sketch[1]
opt = settings.analysis
controller = Controller(analysis_sketch)
from xframe.presenters.matplolibPresenter import heat2D,heat2D_multi

ccd = db.load('ccd')
cc = ccd['cross_correlation']

new_cc,_locals =ccd_remove_0_order(cc)
#locals().update(_locals)
metadata = {}
proj_opt = settings.analysis.projections.reciprocal.pre_init.cross_correlation
max_order = settings.analysis.projections.reciprocal.pre_init.max_order
metadata = {**ccd,'orders':np.arange(max_order+1),'mask':proj_opt.mask,'n_particles':proj_opt.n_particles}        
new_bl = cross_correlation_to_deg2_invariant(new_cc,3,metadata)[0]
bl = cross_correlation_to_deg2_invariant(cc,3,metadata)[0]


ls = 0,2,4,6,8
db.save('/gpfs/exfel/theory_group/user/berberic/MTIP/test/cc_remove_zero_component/cc0.matplotlib',heat2D_multi.get_fig([[bl[l] for l in ls],[new_bl[l] for l in ls]],scale = 'log',shape = [2,len(ls)]))



import ctypes as C 
from ctypes import c_uint,c_double,c_int,c_size_t,c_char,c_void_p
import logging

log=logging.getLogger('root')

#error handler#
gsl_error_handler_t = C.CFUNCTYPE(c_void_p, c_char,c_char,c_int,c_int)

#general Complex#
gsl_complex_data=c_double*2
class gsl_complex(C.Structure):
    _fields_ = [("dat",gsl_complex_data)]



gsl_complex_packed_array = C.POINTER(c_double)


#FFT#
class gsl_fft_real_wavetable(C.Structure):
    _fields_=[
        ("n",c_size_t),
        ("nf",c_size_t),
        ("factor",c_size_t*64),
        ("twiddle",C.POINTER(gsl_complex)*64),
        ("trig",C.POINTER(gsl_complex))
    ]

class gsl_fft_complex_wavetable(C.Structure):
    _fields_=[
        ("n",c_size_t),
        ("nf",c_size_t),
        ("factor",c_size_t*64),
        ("twiddle",C.POINTER(gsl_complex)*64),
        ("trig",C.POINTER(gsl_complex))
    ]

class gsl_fft_halfcomplex_wavetable(C.Structure):
    _fields_=[
        ("n",c_size_t),
        ("nf",c_size_t),
        ("factor",c_size_t*64),
        ("twiddle",C.POINTER(gsl_complex)*64),
        ("trig",C.POINTER(gsl_complex))
    ]

class gsl_fft_real_workspace(C.Structure):
    _fields_=[
        ("n",c_size_t),
        ("scratch",C.POINTER(c_double)),
    ]

class gsl_fft_complex_workspace(C.Structure):
    _fields_=[
        ("n",c_size_t),
        ("scratch",C.POINTER(c_double)),
    ]
    

#DHT#
class gsl_dht_struct(C.Structure):
    _fields_=[
        ('size',c_size_t), #size of sample arrays to be transformed
        ('nu',c_double), #bessel funciton order
        ('xmax',c_double), # realspace cutoff        
        ('kmax',c_double), # reciprocal space cutoff
        ('j',C.POINTER(c_double)), # array of computed j_nu zeros
        ('Jjj',C.POINTER(c_double)), # transform numerator
        ('J2',C.POINTER(c_double))  # transform denominator
    ]



from xframe.startup_routines import DefaultProjectArgparser
from xframe.startup_routines import DefaultProjectClick

class ProjectArgparser(DefaultProjectArgparser):
    def __init__(self):
        super().__init__()
        self.project_description = 'How to setup a project in xFrame'




class ProjectClick(DefaultProjectClick):
    def __init__(self):
        super().__init__()
        self.project_description = 'A short tutorial explaining how to setup a project in xFrame. With details on Settings and File handling as well as multiprocessing and GPU access.'
        self.short_description = 'Example project in xFrame'



import numpy as np
import os
os.chdir('/gpfs/exfel/theory_group/user/berberic/Programs/xframe')
from xframe.database.euxfel import default_DB
from xframe.library.gridLibrary import GridFactory
from xframe.startup_routines import dependency_injection
dependency_injection()
from xframe.presenters.matplolibPresenter import plot1D,heat2D_multi
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward
from xframe.library.mathLibrary import circularHarmonicTransform_complex_forward

db = default_DB()

ccd_1 = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_covid_1p_100k.h5')
#ccd_1 = db.load('/gpfs/exfel/theory_group/user/berberic/MTIP/ccd/3d_covid_1p_100k_small.h5')

cc= ccd_1['intra']['ccf_2p_q1q2']
ccm = np.moveaxis(circularHarmonicTransform_real_forward(cc),2,0)
radial_points = ccd_1['q1']

base_path = '/gpfs/exfel/theory_group/user/berberic/MTIP/test/ccn/'
options = {'plot_range':[1e7,1e16]}

def _save_first_invariants(bls,radial_points,base_path,options,name='model_0_'):
    max_value = np.max(tuple(np.abs(bl).max() for bl in bls))
    vmin,vmax = options.get('plot_range',[max_value*1e-12,max_value])
    if [vmin,vmax] == [None,None]:
        vmin,vmax = [max_value*1e-12,max_value]
        
    grid = GridFactory.construct_grid('uniform',[radial_points,np.arange(bls.shape[2])])
    order_ids = [[2,4,16,24,34],[36,38,40,42,44]]
    #order_ids = [20,40,60,80,127]
    
    shape = [2,len(order_ids[0])]
    layouts = []
    for i in range(shape[0]):
        layout_part = []
        orders = np.array(order_ids[i])
        #orders = np.arange(10*i,10*i+len(order_ids)*2,2)
        for o in orders:
            layout = {'title':'$C_{'+'{}'.format(o)+'}$',
                      'x_label':'$q_1$',
                      'y_label':'$q_2$'
                      }
            layout_part.append(layout)
        layouts.append(layout_part)
                                                
    fig_bl_masks = heat2D_multi.get_fig([[np.abs(bls[n]) for n in order_ids[0]],[np.abs(bls[n]) for n in order_ids[1]] ],scale = 'log',layout = layouts,grid =grid[:],shape = shape,size = (30,10),vmin= vmin, vmax = vmax,cmap='plasma')
    bl_path = base_path +name+'ccn.matplotlib'
    db.save(bl_path,fig_bl_masks,dpi = 300)
    
_save_first_invariants(ccm,radial_points,base_path,options,'covid_100k_2')



import numpy as np
from xframe.library.mathLibrary import circularHarmonicTransform_real_forward as ht
from xframe.library.mathLibrary import circularHarmonicTransform_real_inverse as iht

def generate_calc_cross_correlation(data_shape,pixels_per_radial_ring,mask_type='q_dependend'):    
    n_radial_points = data_shape[0]
    n_angular_points = data_shape[1]
    n_orders = n_angular_points//2 + 1
    coeff_mask = np.ones((n_radial_points,n_orders),dtype=bool)
    inverse_coeff_mask = ~coeff_mask
    orders_per_radial_ring = pixels_per_radial_ring//2 + 1
    for mask_part,max_order in zip(coeff_mask, orders_per_radial_ring):
        mask_part[max_order + 1:]=False
        
    if mask_type == 'q1q2_dependend':        
        def masked_cc(data,mask):
            #assumes mask(n,q1,q2,phi) = mask(n,q2,q1,phi) => mask(n,q1,q2,phi) * mask(n,q2,q1,phi) = mask(n,q1,q2,phi)
            data_harm_coeff = np.zeros(mask.shape,dtype=complex)
            mask_harm_coeff = np.array(data_harm_coeff)
            data_harm_coeff[...,coeff_mask]= ht(data[:,None,...]*mask)[...,coeff_mask]
            data_harm_coeff[...,coeff_mask]= ht(mask)[...,coeff_mask]
            data_cc_coeff = data_harm_coeff*np.swapaxes(data_harm_coeff,1,2).conj()
            mask_cc_coeff = mask_harm_coeff*mask_harm_coeff.conj()
            data_cc = iht(data_cc_coeff,n_angular_points)
            mask_cc = iht(mask_cc_coeff,n_angular_points)
            cc = data_cc/mask_cc
            return cc 
    elif mask_type == 'q_dependend':
        def masked_cc(data,mask):
            data_harm_coeff = np.zeros(mask.shape,dtype=complex)
            mask_harm_coeff = np.array(data_harm_coeff)
            data_harm_coeff[...,coeff_mask]= ht(data*mask)[...,coeff_mask]
            data_harm_coeff[...,coeff_mask]= ht(mask)[...,coeff_mask]
            data_cc_coeff = data_harm_coeff[:,:,None,:]*(data_harm_coeff.conj())[:,None,:,:]
            mask_cc_coeff = mask_harm_coeff[:,:,None,:]*(mask_harm_coeff.conj())[:,None,:,:]
            data_cc = iht(data_cc_coeff,n_angular_points)
            mask_cc = iht(mask_cc_coeff,n_angular_points)
            cc = data_cc/mask_cc
            return cc 
    elif mask_type == 'unmasked':
        def masked_cc(data):
            data_harm_coeff = ht(data)
            data_harm_coeff[...,inverse_coeff_mask]=0
            data_cc_coeff = data_harm_coeff[...,None,:]*(data_harm_coeff.conj())[...,None,:,:]
            cc = iht(data_cc_coeff,n_angular_points)
            return cc 
    return masked_cc



import numpy as np
import logging
log=logging.getLogger('root')

from xframe.library.mathLibrary import SphericalIntegrator,PolarIntegrator,RadialIntegrator


###### FSC/FRC #######
def FSC_bit_limit(target_SNR,grid):
    # calculates FRC/FSC limit (1/2 bit limit is given by target_SNR=1/2)
    # assumes uniform spherical/polar datagrid
    # I am following  DOI: 10.1016/j.jsb.2005.05.009
    # Title: Fourier shell correlation threshold criteria  Authors: Marin van Heel, Michael Schatz
    # See equations (14) and (17) for target_SNR = 1 and 1/2 cases

    points_per_shell = np.prod(grid.shape[1:])
    half_dataset_SNR = target_SNR/2
    x = half_dataset_SNR
    p = points_per_shell
    fsc_limit = (x+2*np.sqrt(x/p)+1/np.sqrt(p))/(1+x+2**np.sqrt(x/p))
    return fsc_limit

def _fsc(a1,a2):
    #Computes FSC/FRC assuming the input scattering_amplitudes are sampled on a uniform spherical/polar grid
    axes = tuple(range(1,a1.ndim))
    I1 = (a1*a1.conj()).real
    I2 = (a2*a2.conj()).real
    nominator = np.sum(a1*a2.conj(),axis = axes).real
    denominator = np.sqrt(np.sum(I1,axis=axes)*np.sum(I2,axis=axes)).real
    non_zero = denominator!=0
    fsc = np.ones(nominator.shape,dtype = float)
    fsc[non_zero]=nominator[non_zero]/denominator[non_zero]
    return fsc

def _fsc_single(a,b):
    #Computes FSC/FRC assuming the input scattering_amplitudes are sampled on a uniform spherical/polar grid
    axes = tuple(range(1,a1.ndim))
    nominator = np.sum(a*a.conj(),axis = axes).real
    denominator = np.sum(b*b.conj(),axis=axes).real
    non_zero = denominator!=0
    fsc = np.ones(a.shape,dtype = float)
    fsc[non_zero]=nominator[non_zero]/denominator[non_zero]
    return fsc

def FSC_two_data_halves(averaged_scattering_amplitudes):
    a1,a2 = averaged_scattering_amplitudes
    return _fsc(a1,a2)
def FSC_single_classical(averaged_scattering_amplitude,input_intensity):
    a = averaged_scattering_amplitude
    b = np.sqrt(input_intensity)
    return _fsc_single(a,b)
def FSC_single_fxs(averaged_scattering_amplitude,averaged_projected_scattering_amplitude):
    a = averaged_scattering_amplitude
    b = averaged_projected_scattering_amplitude
    return _fsc(a,b)


###### PRTF ######
# The classical definitions follows
# Paper: High-resolution ab initio three-dimensional x-ray diffraction microscopy (equation (23))
# Authors: Henry N. Chapman, et. al.
# DOI: https://doi.org/10.1364/JOSAA.23.001179
def PRTF(a1,a2,b1,b2,return_prtf_nd=False):
    axes = tuple(range(1,a1.ndim))
    prtf_nd = np.ones(a1.shape,dtype = complex)
    non_zero = (b1!=0) & (b2!=0)
    prtf_nd[non_zero] = (a1[non_zero]*a2[non_zero].conj())/(b1[non_zero]*b2[non_zero].conj())
    #log.info(np.average(prtf_nd,axis = axes ))
    prtf_nd[~non_zero & (a1!=0) & (a2!=0)]=0
    #log.info('after zero = {}'.format(np.average(prtf_nd,axis = axes )))
    prtf_nd = np.sqrt(prtf_nd)
    #log.info('after sqrt = {}'.format(np.average(prtf_nd,axis = axes )))
    prtf = np.average(prtf_nd,axis = axes )
    #log.info('prtf = {}'.format(prtf))
    prtf_std = np.std(prtf_nd,axis = axes )
    if return_prtf_nd:
        return prtf,prtf_std,prtf_nd
    else:
        return prtf,prtf_std

def PRTF_clasical(averaged_scattering_amplitude,input_intensity,return_prtf_nd=False):
    a = averaged_scattering_amplitude
    b = np.sqrt(input_intensity)
    return PRTF(a,a,b,b,return_prtf_nd=return_prtf_nd)
def PRTF_classical_two_data_halves(averaged_scattering_amplitudes,input_intensities,return_prtf_nd=False):
    a1,a2 = averaged_scattering_amplitudes
    b1,b2 = np.sqrt(input_intensity[0]),np.sqrt(input_intensity[1])
    return PRTF(a1,a2,b1,b2,return_prtf_nd=return_prtf_nd)

def PRTF_fxs(averaged_scattering_amplitude,averaged_intensity,averaged_projected_scattering_amplitude=False, averaged_projected_intensity=False,return_prtf_nd=False):
    if isinstance(averaged_projected_scattering_amplitude,np.ndarray) and isinstance(averaged_projected_intensity,np.ndarray):
        a1 = averaged_scattering_amplitude
        a2 = averaged_projected_scattering_amplitude
        b1 = np.sqrt(averaged_intensity)
        b2 = np.sqrt(averaged_projected_intensity)
        prtf_data = PRTF(a1,a2,b1,b2,return_prtf_nd=return_prtf_nd)
    else:
        a = averaged_scattering_amplitude
        b = np.sqrt(averaged_intensity)
        prtf_data = PRTF(a,a,b,b,return_prtf_nd=return_prtf_nd)
    return prtf_data

###### FQC ######
# The classical definitions follows (with a small change)
# Paper: Correlations in Scattered X-Ray Laser Pulses Reveal Nanoscale Structural Features of Viruses (equation (17) in Supplements)
# Authors: Ruslan P. Kurta, et. al.
# DOI: https://doi.org/10.1103/PhysRevLett.119.158102
# The change is that we leave out the zero's order harmonic coefficient in the summs


def FQC(cn1,cn2,return_cc=False,skip_odd_orders=False):
    #Calculates FQC based on harmonic coefficients of two input cross-correlations
    # Assumes format (q1,q1,harmonic_order)
    # Assumes that cn[...,0] contins the zero'th order coefficient
    if skip_odd_orders:        
        o_start=2
        o_step=2
    else:
        o_start=1
        o_step=1
        
    # symmetrize & select relevant orders
    c1 = ((cn1 + np.swapaxes(cn1,0,1))/2)[...,o_start::o_step]
    c2 = ((cn2 + np.swapaxes(cn2,0,1))/2)[...,o_start::o_step]

    # squares
    C1 = (c1*c1.conj()).real
    C2 = (c2*c2.conj()).real
    
    
    cc_nominator = np.sum(c1*c2.conj(),axis=-1).real
    cc_denominator = np.sqrt(np.sum(C1,axis = -1)*np.sum(C2,axis = -1)).real
    non_zero = cc_denominator != 0

    cc = np.ones(c1.shape[:2],dtype = float)
    cc[non_zero]=cc_nominator[non_zero]/cc_denominator[non_zero]

    qq_mask = np.tril(np.ones(cc.shape)).astype(bool)

    fqc = np.mean(cc,axis = 1,where=qq_mask)
    if return_cc:
        return fqc,cc
    else:
        return fqc

def FQCB_2D(bn1,bn2,return_2d_fqcb=False,skip_odd_orders=False,include_zero_order=False):
    #Calculates FQC based on harmonic coefficients of two input cross-correlations
    # Assumes format (q1,q1,harmonic_order)
    # Assumes that cn[...,0] contins the zero'th order coefficient
    
    if skip_odd_orders:        
        o_start=2
        o_step=2
    else:
        o_start=1
        o_step=1
    if include_zero_order:
        o_start = 0 
    o_stop=min(len(bn1),len(bn2))
        
    # symmetrize & select relevant orders
    b1 = ((bn1 + np.swapaxes(bn1,-1,-2))/2)[o_start:o_stop:o_step]
    b2 = ((bn2 + np.swapaxes(bn2,-1,-2))/2)[o_start:o_stop:o_step]
    #b1 = bn1[o_start:o_stop:o_step]
    #b2 = bn2[o_start:o_stop:o_step]
    
    # squares
    B1 = (b1*b1.conj()).real
    B2 = (b2*b2.conj()).real
    
    bb_nominator = np.sum(b1*b2.conj(),axis = 0).real
    bb_denominator = np.sqrt(np.sum(B1,axis = 0)*np.sum(B2,axis = 0)).real
    non_zero = bb_denominator != 0

    bb = np.ones(b1.shape[1:],dtype = float)
    #log.info('bb shape = {} nonzero shape {} nom shape ={} denom shape = {}'.format(bb.shape,non_zero.shape,bb_nominator.shape,bb_denominator.shape))
    bb[non_zero]=bb_nominator[non_zero]/bb_denominator[non_zero]
    bb = np.abs(bb)

    qq_mask = np.tril(np.ones(bb.shape)).astype(bool)

    fqc = np.mean(bb,axis = 1,where=qq_mask)
    std = np.std(bb,axis=1,where=qq_mask)
    if return_2d_fqcb:
        return fqc,std,bb
    else:
        return fqc,std
    
def FQCB_3D(bn1,bn2,return_2d_fqcb=False,skip_odd_orders=False,include_zero_order=False):
    #Calculates FQC based on harmonic coefficients of two input cross-correlations
    # Assumes format (q1,q1,harmonic_order)
    # Assumes that cn[...,0] contins the zero'th order coefficient
    raise NotImplementedError()
    if skip_odd_orders:        
        o_start=2
        o_step=2
    else:
        o_start=1
        o_step=1
    if include_zero_order:
        o_start = 0 
    o_stop=min(len(bn1),len(bn2))
        
    # symmetrize & select relevant orders
    b1 = ((bn1 + np.swapaxes(bn1,-1,-2))/2)[o_start:o_stop:o_step]
    b2 = ((bn2 + np.swapaxes(bn2,-1,-2))/2)[o_start:o_stop:o_step]
    #b1 = bn1[o_start:o_stop:o_step]
    #b2 = bn2[o_start:o_stop:o_step]
    
    # squares
    B1 = (b1*b1.conj()).real
    B2 = (b2*b2.conj()).real
    
    bb_nominator = np.sum(b1*b2.conj(),axis = 0).real
    bb_denominator = np.sqrt(np.sum(B1,axis = 0)*np.sum(B2,axis = 0)).real
    non_zero = bb_denominator != 0

    bb = np.ones(b1.shape[1:],dtype = float)
    #log.info('bb shape = {} nonzero shape {} nom shape ={} denom shape = {}'.format(bb.shape,non_zero.shape,bb_nominator.shape,bb_denominator.shape))
    bb[non_zero]=bb_nominator[non_zero]/bb_denominator[non_zero]
    bb = np.abs(bb)

    qq_mask = np.tril(np.ones(bb.shape)).astype(bool)

    fqc = np.mean(bb,axis = 1,where=qq_mask)
    std = np.std(bb,axis=1,where=qq_mask)
    if return_2d_fqcb:
        return fqc,std,bb
    else:
        return fqc,std



import numpy as np
from xframe.library.gridLibrary import NestedArray,ReGrider
from xframe.presenters.matplolibPresenter import plot1D

data = np.arange(100)*np.pi/50

g_a = NestedArray((np.arange(100)/4)[:,None],1)
g_b = NestedArray((np.arange(300)/2)[:,None],1)

new_data = ReGrider.regrid(data,g_a,'cartesian',g_b,'cartesian',options = {'interpolation':'linear','fill_value':0.0})



import traceback
import os
import sys
import xframe
import logging
from importlib import util
log = logging.getLogger('root')

experiments_to_load = xframe.settings.general.load_experiments
def create_init_file_if_non_existing(experiment_path):
    init_path = os.path.join(experiment_path,'__init__.py')    
    if not os.path.exists(init_path):
        with open(init_path, 'w') as fp:
            fp.write('')
    
def import_experiment(name,path):
    create_init_file_if_non_existing(path)
    py_import_path = f'xframe.experiments.{name}'
    spec = util.spec_from_file_location(py_import_path, path+'__init__.py')
    module = util.module_from_spec(spec)
    sys.modules[py_import_path] = module
    spec.loader.exec_module(module)
    globals().update({name:module})

for experiment_name,experiment_path in xframe.known_experiments.items():
    try:
        if isinstance(experiments_to_load,(list,tuple)):
            if experiment_name in experiments_to_load:
                import_experiment(experiment_name,experiment_path)
        else:        
            import_experiment(experiment_name,experiment_path)
    except Exception as e:
        log.error(f'Error during import of experiment at {experiment_path}, Traceback:\n'+traceback.format_exc())



import sys
import h5py
import numpy as np
import glob
import time
import logging
import traceback
log=logging.getLogger('root')

import ctypes
from psutil import virtual_memory
from scipy import ndimage


from xframe.experiment.interfaces import CalibratorInterface
from xframe.library import pythonLibrary as pyLib
from xframe import database
from xframe import settings

class MockCalibrator(CalibratorInterface):
    def calibration_worker(self,data_generator,out_modifier = False):
        pass
    
class AGIPD_VDS_Calibrator(CalibratorInterface):
    '''
    Interface to get frames interactively
    Initially specify path to folder with raw/proc data
    Then use get_frame(num) to get specific frame
    '''
    def __init__(self):
        #            geom_fname='/gpfs/exfel/exp/SPB/201802/p002145/scratch/geom/b1.geom'):
        #log.info(opt)
        data_mode = settings.experiment.data_mode
        self.good_cells = settings.experiment.good_cells
        opt = settings.experiment.get('calibrator',{})
        self.db=database.experiment
        self.n_modules = 16
        self.verbose = opt.get('verbose',0)
        self.sum_h5cells = len(self.good_cells)
        #        self.geom_fname = opt.get('geom_fname','../geometry/b2.geom')
        self.mask_lg_mg = opt.get('mask_low_and_mid_gain',False)
        self.dset_name = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/data'
        self.train_name = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/trainId'
        self.pulse_name = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/pulseId'
        self.cell_name = '/INSTRUMENT/SPB_DET_AGIPD1M-1/DET/image/cellId'
        if data_mode=='raw':
            self.calib = self.db.load('calibration_constants')
            self.cmode = opt.get('common_mode_correction',False)
            log.info(opt.keys())
            log.info('common mode = {}'.format(self.cmode))
            self.photon_threshold= opt.get('photon_threshold',0.5)
            self.gain_levels=self.calib['DigitalGainLevel']#np.concatenate((([self.calib[module]['DigitalGainLevel'] for module in range(16)],dtype=np.float32)
            self.analog_offset=self.calib['AnalogOffset']#np.array([self.calib[module]['AnalogOffset'] for module in range(16)],dtype=np.float32)
            self.relative_gain=self.calib['RelativeGain']#np.array([self.calib[module]['RelativeGain'] for module in range(16)],dtype=np.float32)
            badpix=self.calib['Badpixel']
            if self.mask_lg_mg:
                badpix[(self.gain_levels[1] | self.gain_levels[2])]=1 # mask all pixels in MG and LG, with some positive value    # notice , all masked values are >0, all good values ==0 in the mask    
            self.bad_pixel=badpix#np.array([self.calib[module]['Badpixel'] for module in range(16)])        
        
        self.frame = np.empty((16,512,128))
        self.frame_shape=(512,128)
        
        self.ids_by_run = {}
        self.cell_ids = None
        self.pulse_ids = None
        self.train_ids = None
        self.nframes = None

        log.info("opt keys ={}".format(opt.keys()))
        self.gain_mode = opt.get('gain_mode','adaptive')
        
        
        # bd.load(calibration)  [h5py.File(f, 'r') for f in sorted(glob.glob(calib_glob))] '/gpfs/exfel/exp/SPB/201802/p002145/scratch/calib/r%.4d/Cheetah*.h5'
        #dir_calib = '/gpfs/exfel/u/scratch/SPB/201802/p002145/kurta/calibData/'
        #calib_str=dir_calib+'Cheetah-AGIPD-calib.h5'
        #mask_str=dir_calib+'bad-pixel-mask.h5'
        
    def set_ids_for_run(self,run):
        ids_by_run = self.ids_by_run
        run_str=str(run)
        if run_str in ids_by_run:
            self.cell_ids = ids_by_run[run_str]['cell_ids']
            self.pulse_ids = ids_by_run[run_str]['pulse_ids']
            self.train_ids = ids_by_run[run_str]['train_ids']
        else:
            load = self.db.load
            pm = {'run':run,'data_mode':'raw'}
            cell_ids = load('cell_ids',path_modifiers = pm)
            pulse_ids = load('pulse_ids',path_modifiers = pm)
            train_ids = load('train_ids',path_modifiers = pm)
            self.ids_by_run[run_str] = {'cell_ids':cell_ids,'pulse_ids':pulse_ids, 'train_ids':train_ids}
            self.cell_ids = cell_ids
            self.pulse_ids = pulse_ids
            self.train_ids = train_ids
        self.nframes=len(self.cell_ids)
            
    def load_calib_constants(self):
        calib_mode = self.calib_mode.lower()
        calib_glob = None
        calib ={}
        if calib_mode == 'cheetah':
            if self.calib_run is None:
                run_str='latest'
            else:
                run_str= 'r%.4d' % calib_run
            f = self.db.load('calibration', opt = {'mode':'cheetah'}, path_modifiers={'run':run_str})
            #(gain,cell,module,many_pixel_ax,few_pixel_ax)
            calib['DigitalGainLevel'] = np.concatenate(tuple(f[module]['DigitalGainLevel'][:][:,:,None,...] for module in range(self.n_modules)),axis = 2,dtype = np.float32)
            calib['AnalogOffset']= np.concatenate(tuple(f[module]['AnalogOffset'][:][:,:,None,...] for module in range(self.n_modules)),axis = 2,dtype = np.float32)
            calib['RelativeGain'] = np.concatenate(tuple(f[module]['RelativeGain'][:][:,:,None,...] for module in range(self.n_modules)),axis = 2,dtype = np.float32)
            calib['Badpixel'] = np.concatenate(tuple(f[module]['Badpixel'][:][:,:,None,...] for module in range(self.n_modules)),axis = 2)
        elif calib_mode == 'alireza2020ver1':
            calib_data = self.db.load('calibration',opt = {'mode':'alireza'})
            data = calib_data['data']
            masks = calib_data['masks']
            npls = self.sum_h5cells
            calib = {}
            npls = 1#data['AnalogOffset'][:]
            calib['AnalogOffset']=data['AnalogOffset'][:].reshape(3,npls,16,512,128) 
            calib['RelativeGain']=data['RelativeGain'][:].reshape(3,npls,16,512,128) 
            gain_modes=np.asarray(data['DigitalGainLevel']).reshape(3,npls,16,512,128)
            calib['DigitalGainLevel'] = gain_modes
            
    
            calib['Badpixel']=np.asarray(masks['badPixelMask']).reshape(3,npls,16,512,128) # mask from dark run Analysis
            print('Mask generated from dark runs analysis will be applied')
            if 'badPixelMask_ByWater' in masks:
                calib['Badpixel']=np.add(calib['Badpixel'], masks['badPixelMask_ByWater'][:].reshape(3,npls,16,512,128)) # mask from Water analysis
                print('Mask generated from water data analysis will be applied')
            if self.mask_lg_mg:
                badpix[(gain_modes[1] | gain_modes[2])]=1 # mask all pixels in MG and LG, with some positive value    # notice , all masked values are >0, all good values ==0 in the mask    
        else:
            print('Error: Calibration Mode {} not known.'.format(calib_mode))
        return calib


    def _calibrate_frames_adaptive_new(self,data_file,frames,module):
        nan_ids=np.isnan(data_file).nonzero()
        if len(nan_ids) != 0:
            #print('nans in datafile to -5')
            data_file[nan_ids]=-5
            
        cmode = self.cmode
        photonThresh = self.photon_threshold
        t_len=len(data_file)
        n_frames=len(frames)
        gains=data_file[:,1]
        raw_data=data_file[:,0]
        cells=self.cell_ids[frames]
        
        
        no_data_mask=(cells==65535)
        raw_data=raw_data[~no_data_mask]
        gains=gains[~no_data_mask]
        cells=cells[~no_data_mask]
        #print('data2 shape={}'.format(data.shape))
        #        print('cells={}'.format(frames))
        gain_modes = self._threshold2(gains, module, cells)
        del(gains)
        offset = np.empty(gain_modes.shape[1:])
        gain = np.empty(gain_modes.shape[1:])
        badpix = np.empty(gain_modes.shape[1:])
        for gain_level,gain_mask in enumerate(gain_modes):
            #print('analog_offset shape={}'.format(self.analog_offset[module,gain_level,cells].shape))
            offset[gain_mask] = self.analog_offset[gain_level,cells,module][gain_mask]
            gain[gain_mask] = self.relative_gain[gain_level,cells,module][gain_mask]
            badpix[gain_mask] = self.bad_pixel[gain_level,cells,module][gain_mask]
            #        print('data dtype= {} offset dtype={} gain dtype={}'.format(data.dtype,offset.dtype,gain.dtype))
        #data_and_mask = np.zeros_like(data_file,dtype = np.float32)
        data = np.zeros_like(data_file[:,0],dtype=np.float32)
        mask = np.full(data_file[:,1].shape,True)
        #data = data_and_mask[:,0]
        #mask = data_and_mask[:,1]
        #mask[:]=1
        data[:] = (np.float32(raw_data) - offset)*gain
        del(raw_data)
        del(data_file)

        #assumes badpix is True for bad pixels and reverses this convention
        bad_pixel_mask = ~(badpix != 0)
        mask[~bad_pixel_mask] = False
        data[~bad_pixel_mask] = 0
        del(bad_pixel_mask)
        del(badpix)
        if self.verbose > 1:
            print('Found %d bad pixels for module %d' % ((badpix != 0).sum(), module))
            
        # Threshold below 0.5-0.7 photon (1 photon = 45 ADU)            
        if isinstance(photonThresh,(list,tuple)):
            photon_threshold_mask = (data < photonThresh[0]*45*gain) | (data>photonThresh[1]*45*gain)
        elif isinstance(photonThresh,(int,float)):
            photon_threshold_mask = (data < photonThresh*45*gain)
                        
        data[photon_threshold_mask] = 0
        #mask[photon_threshold_mask] = 1        
        
        if cmode:
            # Median subtraction by 64x64
            #log.info('applying common mode correction!')
            tmp_data = data.reshape(n_frames,8,64,2,64).transpose(0,2,4,1,3).reshape(n_frames,64,64,16)
            if self.verbose > 1:
                print('Common-mode correction for module %d: %.1f ADU' % (module, np.sum(np.median(data, axis=(0,1)))))
#            print('new data shape={}'.format(data.shape))
#            print('median shape ={}'.format(np.median(data,axis=(1,2)).shape))
            tmp_data -= np.median(tmp_data, axis=(1,2))[:,None,None,:]
            tmp_data = tmp_data.reshape(n_frames,64,64,8,2).transpose(0,3,1,4,2).reshape(n_frames,512,128)
            data[:]=tmp_data
#        print('data shape={}'.format(data.shape))

        
            #print('Setting %d pixels below photon threshold to zero for module %d' % (((data < photonThresh*45*gain) & (data > 0)).sum(), module))
        #data[data < photonThresh*45*gain] = 0
        #data[data > 10000] = 10000
        #print('data shape={}'.format(data.shape))
        return data, mask
    def _calibrate_frames_fixed_gain_medium_new(self,data_file,frames,module): 
        nan_ids=np.isnan(data_file).nonzero()
        if len(nan_ids) != 0:
            #print('nans in datafile to -5')
            data_file[nan_ids]=-5
            
        cmode = self.cmode
        photonThresh = self.photon_threshold
        t_len=len(data_file)
        n_frames=len(frames)
        #gains=data_file[:,1]
        raw_data=data_file[:,0]
        cells=self.cell_ids[frames]
        
        
        no_data_mask=(cells==65535)
        raw_data=raw_data[~no_data_mask]
        #gains=gains[~no_data_mask]
        cells=cells[~no_data_mask]
        #print('data2 shape={}'.format(data.shape))
        #        print('cells={}'.format(frames))
        #ain_modes = self._threshold2(gains, module, cells)
        #del(gains)
        offset = self.analog_offset[1,cells,module]
        gain = self.relative_gain[1,cells,module]
        badpix = self.bad_pixel[1,cells,module]

        data = np.zeros_like(data_file[:,0],dtype=np.float32)
        mask = np.full(data_file[:,1].shape,True)
        #data_and_mask = np.zeros_like(data_file,dtype = np.float32)
        #data = data_and_mask[:,0]
        #mask = data_and_mask[:,1]
        #mask[:]=1
        data[:] = (np.float32(raw_data) - offset)*gain
        del(raw_data)
        del(data_file)

        #assumes badpix is True for bad pixels and reverses this convention
        bad_pixel_mask = ~(badpix != 0)
        mask[~bad_pixel_mask] = False
        data[~bad_pixel_mask] = 0
        del(bad_pixel_mask)
        del(badpix)
        if self.verbose > 1:
            print('Found %d bad pixels for module %d' % ((badpix != 0).sum(), module))
            
        # Threshold below 0.5-0.7 photon (1 photon = 45 ADU)            
        if isinstance(photonThresh,(list,tuple)):
            photon_threshold_mask = (data < photonThresh[0]*45*gain) | (data>photonThresh[1]*45*gain)
        elif isinstance(photonThresh,(int,float)):
            photon_threshold_mask = (data < photonThresh*45*gain)
                        
        data[photon_threshold_mask] = 0
        #mask[photon_threshold_mask] = 1        
        
        if cmode:
            # Median subtraction by 64x64
            #log.info('applying common mode correction!')
            tmp_data = data.reshape(n_frames,8,64,2,64).transpose(0,2,4,1,3).reshape(n_frames,64,64,16)
            if self.verbose > 1:
                print('Common-mode correction for module %d: %.1f ADU' % (module, np.sum(np.median(data, axis=(0,1)))))
#            print('new data shape={}'.format(data.shape))
#            print('median shape ={}'.format(np.median(data,axis=(1,2)).shape))
            tmp_data -= np.median(tmp_data, axis=(1,2))[:,None,None,:]
            tmp_data = tmp_data.reshape(n_frames,64,64,8,2).transpose(0,3,1,4,2).reshape(n_frames,512,128)
            data[:]=tmp_data
#        print('data shape={}'.format(data.shape))

        
            #print('Setting %d pixels below photon threshold to zero for module %d' % (((data < photonThresh*45*gain) & (data > 0)).sum(), module))
        #data[data < photonThresh*45*gain] = 0
        #data[data > 10000] = 10000
        #print('data shape={}'.format(data.shape))
        return data,mask

    def _calibrate_frames_adaptive_old(self,data_file,frames,module):
        nan_ids=np.isnan(data_file).nonzero()
        if len(nan_ids) != 0:
            #print('nans in datafile to -5')
            data_file[nan_ids]=-5
            
        cmode = self.cmode
        photonThresh = self.photon_threshold
        t_len=len(data_file)
        n_frames=len(frames)
        gains=data_file[:,1]
        raw_data=data_file[:,0]
        cells=self.cell_ids[frames]
        
        
        no_data_mask=(cells==65535)
        raw_data=raw_data[~no_data_mask]
        gains=gains[~no_data_mask]
        cells=cells[~no_data_mask]
        #print('data2 shape={}'.format(data.shape))
        #        print('cells={}'.format(frames))
        gain_modes = self._threshold2(gains, module, cells)
        del(gains)
        offset = np.empty(gain_modes.shape[1:])
        gain = np.empty(gain_modes.shape[1:])
        badpix = np.empty(gain_modes.shape[1:])
        for gain_level,gain_mask in enumerate(gain_modes):
            #print('analog_offset shape={}'.format(self.analog_offset[module,gain_level,cells].shape))
            offset[gain_mask] = self.analog_offset[gain_level,cells,module][gain_mask]
            gain[gain_mask] = self.relative_gain[gain_level,cells,module][gain_mask]
            badpix[gain_mask] = self.bad_pixel[gain_level,cells,module][gain_mask]
            #        print('data dtype= {} offset dtype={} gain dtype={}'.format(data.dtype,offset.dtype,gain.dtype))
        data_and_mask = np.zeros_like(data_file,dtype = np.float32)
        data = data_and_mask[:,0]
        mask = data_and_mask[:,1]
        mask[:]=1
        data[:] = (np.float32(raw_data) - offset)*gain
        del(raw_data)
        del(data_file)

        #assumes badpix is True for bad pixels and reverses this convention
        bad_pixel_mask = ~(badpix != 0)
        mask[~bad_pixel_mask] = 0
        data[~bad_pixel_mask] = 0
        del(bad_pixel_mask)
        del(badpix)
        if self.verbose > 1:
            print('Found %d bad pixels for module %d' % ((badpix != 0).sum(), module))
            
        # Threshold below 0.5-0.7 photon (1 photon = 45 ADU)            
        if isinstance(photonThresh,(list,tuple)):
            photon_threshold_mask = (data < photonThresh[0]*45*gain) | (data>photonThresh[1]*45*gain)
        elif isinstance(photonThresh,(int,float)):
            photon_threshold_mask = (data < photonThresh*45*gain)
                        
        data[photon_threshold_mask] = 0
        #mask[photon_threshold_mask] = 1        
        
        if cmode:
            # Median subtraction by 64x64
            #log.info('applying common mode correction!')
            tmp_data = data.reshape(n_frames,8,64,2,64).transpose(0,2,4,1,3).reshape(n_frames,64,64,16)
            if self.verbose > 1:
                print('Common-mode correction for module %d: %.1f ADU' % (module, np.sum(np.median(data, axis=(0,1)))))
#            print('new data shape={}'.format(data.shape))
#            print('median shape ={}'.format(np.median(data,axis=(1,2)).shape))
            tmp_data -= np.median(tmp_data, axis=(1,2))[:,None,None,:]
            tmp_data = tmp_data.reshape(n_frames,64,64,8,2).transpose(0,3,1,4,2).reshape(n_frames,512,128)
            data[:]=tmp_data
#        print('data shape={}'.format(data.shape))

        
            #print('Setting %d pixels below photon threshold to zero for module %d' % (((data < photonThresh*45*gain) & (data > 0)).sum(), module))
        #data[data < photonThresh*45*gain] = 0
        #data[data > 10000] = 10000
        #print('data shape={}'.format(data.shape))
        return data_and_mask
    def _calibrate_frames_fixed_gain_medium_old(self,data_file,frames,module): 
        nan_ids=np.isnan(data_file).nonzero()
        if len(nan_ids) != 0:
            #print('nans in datafile to -5')
            data_file[nan_ids]=-5
            
        cmode = self.cmode
        photonThresh = self.photon_threshold
        t_len=len(data_file)
        n_frames=len(frames)
        #gains=data_file[:,1]
        raw_data=data_file[:,0]
        cells=self.cell_ids[frames]
        
        
        no_data_mask=(cells==65535)
        raw_data=raw_data[~no_data_mask]
        #gains=gains[~no_data_mask]
        cells=cells[~no_data_mask]
        #print('data2 shape={}'.format(data.shape))
        #        print('cells={}'.format(frames))
        #ain_modes = self._threshold2(gains, module, cells)
        del(gains)
        offset = self.analog_offset[1,cells,module]
        gain = self.relative_gain[1,cells,module]
        badpix = self.bad_pixel[1,cells,module]

        data_and_mask = np.zeros_like(data_file,dtype = np.float32)
        data = data_and_mask[:,0]
        mask = data_and_mask[:,1]
        mask[:]=1
        data[:] = (np.float32(raw_data) - offset)*gain
        del(raw_data)
        del(data_file)

        #assumes badpix is True for bad pixels and reverses this convention
        bad_pixel_mask = ~(badpix != 0)
        mask[~bad_pixel_mask] = 0
        data[~bad_pixel_mask] = 0
        del(bad_pixel_mask)
        del(badpix)
        if self.verbose > 1:
            print('Found %d bad pixels for module %d' % ((badpix != 0).sum(), module))
            
        # Threshold below 0.5-0.7 photon (1 photon = 45 ADU)            
        if isinstance(photonThresh,(list,tuple)):
            photon_threshold_mask = (data < photonThresh[0]*45*gain) | (data>photonThresh[1]*45*gain)
        elif isinstance(photonThresh,(int,float)):
            photon_threshold_mask = (data < photonThresh*45*gain)
                        
        data[photon_threshold_mask] = 0
        #mask[photon_threshold_mask] = 1        
        
        if cmode:
            # Median subtraction by 64x64
            #log.info('applying common mode correction!')
            tmp_data = data.reshape(n_frames,8,64,2,64).transpose(0,2,4,1,3).reshape(n_frames,64,64,16)
            if self.verbose > 1:
                print('Common-mode correction for module %d: %.1f ADU' % (module, np.sum(np.median(data, axis=(0,1)))))
#            print('new data shape={}'.format(data.shape))
#            print('median shape ={}'.format(np.median(data,axis=(1,2)).shape))
            tmp_data -= np.median(tmp_data, axis=(1,2))[:,None,None,:]
            tmp_data = tmp_data.reshape(n_frames,64,64,8,2).transpose(0,3,1,4,2).reshape(n_frames,512,128)
            data[:]=tmp_data
#        print('data shape={}'.format(data.shape))

        
            #print('Setting %d pixels below photon threshold to zero for module %d' % (((data < photonThresh*45*gain) & (data > 0)).sum(), module))
        #data[data < photonThresh*45*gain] = 0
        #data[data > 10000] = 10000
        #print('data shape={}'.format(data.shape))
        return data_and_mask
    
    def _calibrate_frames_old(self,data_file,frames,module):
        nan_ids=np.isnan(data_file).nonzero()
        if len(nan_ids) != 0:
            #print('nans in datafile to -5')
            data_file[nan_ids]=-5
            
        cmode = self.cmode
        photonThresh = self.photon_threshold
        t_len=len(data_file)
        n_frames=len(frames)
        gains=data_file[:,1]
        data=data_file[:,0]
        cells=self.cell_ids[frames]
        
        
        no_data_mask=(cells==65535)
        data=data[~no_data_mask]
        gains=gains[~no_data_mask]
        cells=cells[~no_data_mask]
        #print('data2 shape={}'.format(data.shape))
        #        print('cells={}'.format(frames))
        gain_modes = self._threshold2(gains, module, cells)
        offset = np.empty(gain_modes.shape[1:])
        gain = np.empty(gain_modes.shape[1:])
        badpix = np.empty(gain_modes.shape[1:])
        for gain_level,gain_mask in enumerate(gain_modes):
            #print('analog_offset shape={}'.format(self.analog_offset[module,gain_level,cells].shape))
            offset[gain_mask] = self.analog_offset[gain_level,cells,module][gain_mask]
            gain[gain_mask] = self.relative_gain[gain_level,cells,module][gain_mask]
            badpix[gain_mask] = self.bad_pixel[gain_level,cells,module][gain_mask]
#        print('data dtype= {} offset dtype={} gain dtype={}'.format(data.dtype,offset.dtype,gain.dtype))
        data = (np.float32(data) - offset)*gain
        data[badpix != 0] = 0        
        if self.verbose > 1:
            print('Found %d bad pixels for module %d' % ((badpix != 0).sum(), module))    
        if cmode:
            # Median subtraction by 64x64
            #log.info('applying common mode correction!')
            data = data.reshape(n_frames,8,64,2,64).transpose(0,2,4,1,3).reshape(n_frames,64,64,16)
            if self.verbose > 1:
                print('Common-mode correction for module %d: %.1f ADU' % (module, np.sum(np.median(data, axis=(0,1)))))
#            print('new data shape={}'.format(data.shape))
#            print('median shape ={}'.format(np.median(data,axis=(1,2)).shape))
            data -= np.median(data, axis=(1,2))[:,None,None,:]
            data = data.reshape(n_frames,64,64,8,2).transpose(0,3,1,4,2).reshape(n_frames,512,128)
#        print('data shape={}'.format(data.shape))
        # Threshold below 0.5-0.7 photon (1 photon = 45 ADU)

        if isinstance(photonThresh,(list,tuple)):                
            data[(data < photonThresh[0]*45*gain) | (data>photonThresh[1]*45*gain)] = 0
        elif isinstance(photonThresh,(int,float)):
            data[data < photonThresh*45*gain] = 0
        
            #print('Setting %d pixels below photon threshold to zero for module %d' % (((data < photonThresh*45*gain) & (data > 0)).sum(), module))
        #data[data < photonThresh*45*gain] = 0
        #data[data > 10000] = 10000
        #print('data shape={}'.format(data.shape))
        return data
    
    def calibration_worker(self,chunked_data,out_modifier = False):
        
        out_dict={}
        modules,c_id,chunk,n_chunks,data = chunked_data

        start_time = time.time()
        print('calibrating chunk {}/{} for modules {}]'.format(c_id+1,n_chunks,modules))
        for m_id,module in enumerate(modules):
            calib_data_chunk=self._calibrate_frames(data,chunk[m_id],module)
            if callable(out_modifier):
                out_data_part=out_modifier(calib_data_chunk,m_id,module,c_id)
            else:
                out_data_part=calib_data_chunk

            if not str(m_id) in out_dict:
                out_dict[str(m_id)]={str(c_id):out_data_part}
            else:
                out_dict[str(m_id)][str(c_id)] = out_data_part            
        end_time=time.time()
        print('{} seconds per chunk'.format(end_time-start_time))
        n_modules = len(out_dict)
        out_data_per_module = tuple(np.concatenate(tuple(out_dict[str(m_id)].values()),axis = 0)[:,None] for m_id in range(n_modules))
        out_data = np.concatenate(out_data_per_module,axis = 1)
        return out_data

    def calibrate_new(self,data,frame_ids,modules,out_modifier = False):
        gain_mode= self.gain_mode
        try:
            if gain_mode == 'adaptive':
                calibration_routine = self._calibrate_frames_adaptive_new
            elif gain_mode == 'fixed_gain_medium':
                calibration_routine = self._calibrate_frames_fixed_gain_medium_new
            else:
                e = AssertionError('gain mode {} not known'.format(gain_mode))
                raise(e)
        except AssertionError as e:
            log.error(e)
            traceback.print_exc()
            
        out_dict={}
        for m_id,module in enumerate(modules):
            out_dict[str(m_id)]= calibration_routine(data[m_id],frame_ids,module)
        n_modules = len(out_dict)
        out_data_sets_per_module = zip(*tuple([ out_dict[str(m_id)][0][:,None],out_dict[str(m_id)][1][:,None] ] for m_id in range(n_modules)))
        out_data_sets = [np.concatenate(dataset,axis = 1) for dataset in out_data_sets_per_module]
        if callable(out_modifier):
            out_data_sets = out_modifier(out_data_sets,frame_ids,modules)
        return out_data_sets

    def calibrate_old(self,data,frame_ids,modules,out_modifier = False):
        gain_mode= self.gain_mode
        try:
            if gain_mode == 'adaptive':
                calibration_routine = self._calibrate_frames_adaptive_old
            elif gain_mode == 'fixed_gain_medium':
                calibration_routine = self._calibrate_frames_fixed_gain_medium_old
            else:
                e = AssertionError('gain mode {} not known'.format(gain_mode))
                raise(e)
        except AssertionError as e:
            log.error(e)
            traceback.print_exc()
            
        out_dict={}
        for m_id,module in enumerate(modules):
            out_dict[str(m_id)]= calibration_routine(data[m_id],frame_ids,module)
        n_modules = len(out_dict)
        out_data_per_module = tuple(out_dict[str(m_id)][:,None] for m_id in range(n_modules))
        out_data = np.concatenate(out_data_per_module,axis = 1)
        if callable(out_modifier):
            out_data = out_modifier(out_data,frame_ids,modules)
        return out_data

    def _threshold2(self, gains,module,cells):
        thresholds=self.gain_levels[:,cells,module]
        low_gains = gains > thresholds[2,:]
        high_gains = gains < thresholds[1,:]
        medium_gains= (~low_gains)*(~high_gains)
#        print('447,126, low gain={} medium gain {} high gain {}'.format(low_gains[:,447,126],medium_gains[:,447,126],high_gains[:,447,126]))
        return np.concatenate((high_gains[None,...],medium_gains[None,...],low_gains[None,...]),axis=0)

    





import os
file_path = os.path.realpath(__file__)
externalLibraries_dir = os.path.dirname(file_path)
import numpy as np
import ctypes as C 
from ctypes import c_uint,c_double,c_int,c_size_t,c_void_p
import logging

from xframe.library import pythonLibrary as pyLib
from xframe.externalLibraries.gsl_plugin import gslStructures as gslStruct

log=logging.getLogger('root')

cblas = pyLib.load_clib('gslcblas',mode=C.RTLD_GLOBAL)
print("gsl file path = {}".format(file_path))
gsl = pyLib.load_clib('gsl')

try:
    dht = pyLib.load_clib(externalLibraries_dir+'/dht.so')
    dht_found = True
except OSError as e:
    dht_found = False
    log.info('Could not load custom discrete hankel transform Library "dht.so" . Skipping dht')
    
class Wraper:
    def __init__(self):
        def setTypes(function,restype,argtype):
            function.restype=restype
            if isinstance(argtype,list):
                function.argtypes=argtype
            else:
                function.argtype=argtype
            return function


        #error handling#
        gsl_set_error_handler_off=setTypes(gsl.gsl_set_error_handler_off,gslStruct.gsl_error_handler_t,c_void_p)
        
        #defining input and output types for all used gsl functions#
        #FFT mixed Radix#
        #complex
        gsl_fft_complex_wavetable_alloc=gsl.gsl_fft_complex_wavetable_alloc
        gsl_fft_complex_wavetable_alloc.restype=C.POINTER(gslStruct.gsl_fft_complex_wavetable)
        gsl_fft_complex_wavetable_alloc.argtype=c_size_t

        gsl_fft_complex_wavetable_free=gsl.gsl_fft_complex_wavetable_free
        gsl_fft_complex_wavetable_free.restype=None
        gsl_fft_complex_wavetable_free.argtype=C.POINTER(gslStruct.gsl_fft_complex_wavetable)

        gsl_fft_complex_workspace_alloc=gsl.gsl_fft_complex_workspace_alloc
        gsl_fft_complex_workspace_alloc.restype=C.POINTER(gslStruct.gsl_fft_complex_workspace)
        gsl_fft_complex_workspace_alloc.argtype=c_size_t

        gsl_fft_complex_workspace_free=gsl.gsl_fft_complex_workspace_free
        gsl_fft_complex_workspace_free.restype=None
        gsl_fft_complex_workspace_free.argtype=C.POINTER(gslStruct.gsl_fft_complex_workspace)

        gsl_fft_complex_forward=gsl.gsl_fft_complex_forward
        gsl_fft_complex_forward.restype=c_int
        gsl_fft_complex_forward.argtype=[gslStruct.gsl_complex_packed_array, c_size_t, c_size_t, C.POINTER(gslStruct.gsl_fft_complex_wavetable),C.POINTER(gslStruct.gsl_fft_complex_workspace)]

        gsl_fft_complex_inverse=gsl.gsl_fft_complex_inverse
        gsl_fft_complex_inverse.restype=c_int
        gsl_fft_complex_inverse.argtype=[gslStruct.gsl_complex_packed_array, c_size_t, c_size_t, C.POINTER(gslStruct.gsl_fft_complex_wavetable),C.POINTER(gslStruct.gsl_fft_complex_workspace)]

        #real
        gsl_fft_real_wavetable_alloc=gsl.gsl_fft_real_wavetable_alloc
        gsl_fft_real_wavetable_alloc.restype=C.POINTER(gslStruct.gsl_fft_real_wavetable)
        gsl_fft_real_wavetable_alloc.argtype=c_size_t

        gsl_fft_real_wavetable_free=gsl.gsl_fft_real_wavetable_free
        gsl_fft_real_wavetable_free.restype=None
        gsl_fft_real_wavetable_free.argtype=C.POINTER(gslStruct.gsl_fft_real_wavetable)

        gsl_fft_real_workspace_alloc=gsl.gsl_fft_real_workspace_alloc
        gsl_fft_real_workspace_alloc.restype=C.POINTER(gslStruct.gsl_fft_real_workspace)
        gsl_fft_real_workspace_alloc.argtype=c_size_t

        gsl_fft_real_workspace_free=gsl.gsl_fft_real_workspace_free
        gsl_fft_real_workspace_free.restype=None
        gsl_fft_real_workspace_free.argtype=C.POINTER(gslStruct.gsl_fft_real_workspace)

        gsl_fft_real_unpack=gsl.gsl_fft_real_unpack
        gsl_fft_real_unpack.restype=c_int
        gsl_fft_real_unpack.argtype=[C.POINTER(c_double),gslStruct.gsl_complex_packed_array,c_size_t, c_size_t]

        gsl_fft_real_forward=gsl.gsl_fft_real_transform
        gsl_fft_real_forward.restype=c_int
        gsl_fft_real_forward.argtype=[C.POINTER(c_double), c_size_t, c_size_t, C.POINTER(gslStruct.gsl_fft_real_wavetable),C.POINTER(gslStruct.gsl_fft_real_workspace)]
        #halfcomplex
        gsl_fft_halfcomplex_wavetable_alloc=gsl.gsl_fft_halfcomplex_wavetable_alloc
        gsl_fft_halfcomplex_wavetable_alloc.restype=C.POINTER(gslStruct.gsl_fft_halfcomplex_wavetable)
        gsl_fft_halfcomplex_wavetable_alloc.argtype=c_size_t

        gsl_fft_halfcomplex_wavetable_free=gsl.gsl_fft_halfcomplex_wavetable_free
        gsl_fft_halfcomplex_wavetable_free.restype=None
        gsl_fft_halfcomplex_wavetable_free.argtype=C.POINTER(gslStruct.gsl_fft_halfcomplex_wavetable)

        gsl_fft_halfcomplex_unpack=gsl.gsl_fft_halfcomplex_unpack
        gsl_fft_halfcomplex_unpack.restype=c_int
        gsl_fft_halfcomplex_unpack.argtype=[C.POINTER(c_double),C.POINTER(c_double),c_size_t,c_size_t]
        
        gsl_fft_halfcomplex_inverse=gsl.gsl_fft_halfcomplex_inverse
        gsl_fft_halfcomplex_inverse.restype=c_int
        gsl_fft_halfcomplex_inverse.argtype=[C.POINTER(c_double), c_size_t, c_size_t, C.POINTER(gslStruct.gsl_fft_halfcomplex_wavetable),C.POINTER(gslStruct.gsl_fft_real_workspace)]

        
        #BesselZero#
        gsl_sf_bessel_zero_Jnu=gsl.gsl_sf_bessel_zero_Jnu
        gsl_sf_bessel_zero_Jnu.restype=c_double
        gsl_sf_bessel_zero_Jnu.argtyp=[c_double,c_uint]

        #Legendre Polynomials#
        gsl_sf_legendre_Pl=gsl.gsl_sf_legendre_Pl
        gsl_sf_legendre_Pl.restype=c_double
        gsl_sf_legendre_Pl.argtyp=[c_int,c_double]


        #DHT#
        if dht_found:
            gsl_dht_alloc=setTypes(gsl.gsl_dht_alloc,C.POINTER(gslStruct.gsl_dht_struct),c_size_t)
            gsl_dht_init=setTypes(gsl.gsl_dht_init,c_int,[C.POINTER(gslStruct.gsl_dht_struct),c_double,c_double])
            gsl_dht_new=setTypes(gsl.gsl_dht_new,C.POINTER(gslStruct.gsl_dht_struct),[c_size_t,c_double,c_double])
            gsl_dht_free=setTypes(gsl.gsl_dht_free,None,C.POINTER(gslStruct.gsl_dht_struct))
            gsl_dht_apply=setTypes(gsl.gsl_dht_apply,c_int,[C.POINTER(gslStruct.gsl_dht_struct),C.POINTER(c_double),C.POINTER(c_double)])        
            dht_2D=setTypes(dht.dht_2D,c_int,[c_size_t,C.POINTER(C.POINTER(gslStruct.gsl_dht_struct)),np.ctypeslib.ndpointer(np.complex, ndim=2, flags='C'),np.ctypeslib.ndpointer(np.complex, ndim=2, flags='C'),np.ctypeslib.ndpointer(np.complex, ndim=1, flags='C')])
        
            dht_bare=setTypes(dht.dht_bare,c_int,[C.POINTER(C.POINTER(gslStruct.gsl_dht_struct)),np.ctypeslib.ndpointer(np.complex, ndim=1, flags='C'),np.ctypeslib.ndpointer(np.complex, ndim=1, flags='C'),gslStruct.gsl_complex])
            dht_test=setTypes(dht.test,c_int,[c_size_t,c_size_t,np.ctypeslib.ndpointer(np.complex, ndim=2, flags='C')])
        
        
        #error handling#
        self.gsl_set_error_handler_off=gsl_set_error_handler_off
        
        #fft#
        #MixedRadix#
        #complex
        self.gsl_fft_complex_wavetable_alloc=gsl_fft_complex_wavetable_alloc
        self.gsl_fft_complex_wavetable_free=gsl_fft_complex_wavetable_free
        self.gsl_fft_complex_workspace_alloc=gsl_fft_complex_workspace_alloc
        self.gsl_fft_complex_workspace_free=gsl_fft_complex_workspace_free

        self.gsl_fft_complex_forward=gsl_fft_complex_forward
        self.gsl_fft_complex_inverse=gsl_fft_complex_inverse

        #real
        self.gsl_fft_real_wavetable_alloc=gsl_fft_real_wavetable_alloc
        self.gsl_fft_real_wavetable_free=gsl_fft_real_wavetable_free
        self.gsl_fft_real_workspace_alloc=gsl_fft_real_workspace_alloc
        self.gsl_fft_real_workspace_free=gsl_fft_real_workspace_free
        self.gsl_fft_real_unpack=gsl_fft_real_unpack
        
        self.gsl_fft_real_forward=gsl_fft_real_forward
        
        #halfcomplex
        self.gsl_fft_halfcomplex_wavetable_alloc=gsl_fft_halfcomplex_wavetable_alloc
        self.gsl_fft_halfcomplex_wavetable_free=gsl_fft_halfcomplex_wavetable_free
        self.gsl_fft_halfcomplex_unpack=gsl_fft_halfcomplex_unpack
        
        self.gsl_fft_halfcomplex_inverse=gsl_fft_halfcomplex_inverse


        #BesselZero#
        self.gsl_sf_bessel_zero_Jnu=gsl_sf_bessel_zero_Jnu
        #Legendre Polynomials
        self.gsl_sf_legendre_Pl=gsl_sf_legendre_Pl
        if dht_found:
            #DHT#
            self.gsl_dht_alloc=gsl_dht_alloc     #creates DHT object of given <size>
            self.gsl_dht_init=gsl_dht_init         #initalizes the DHT for a given <DHT object> and value of the <Bessel order> and <real cutoff>
            self.gsl_dht_new=gsl_dht_new      # (the two above at the same time) Creates DHT object of given <size> and initializes it for a griven <Bessel order> and  <real cutoff>  
            self.gsl_dht_free=gsl_dht_free      #frees the allocated memory for a DHT object
            self.gsl_dht_apply=gsl_dht_apply  # Applies a given <DHT object> trasform to an given <input array> and stores it in a given <output array>

            #my ajusted versions of DHT#
            self.dht_2D=dht_2D # computes dht of 2D array along the second axis
            self.dht_bare=dht_bare # modified version of gsl_dht_apply without ajustable prefactor
            self.dht_test=dht_test



import numpy as np
from xframe.library import mathLibrary as mLib
from xframe import log
log=log.setup_custom_logger('root','INFO')

#A = np.random.rand(3*20*10).reshape(3,20,10)
A = np.diag(np.exp(-np.arange(0,200)*np.log(1e+6)/100))
b_free = np.arange(0.9,1,0.1/200)[::-1]
b = b_free + np.random.normal(scale = 0.001,size = 200)
b = np.stack((b))


#locals().update(mLib.approximate_tikhonov_parameters(A,b))
lambd = mLib.approximate_tikhonov_parameters(A,b)
x = mLib.optimal_tikhonov_regularization(A,b)
x2 = np.linalg.lstsq(A,b,rcond = None)[0][None,...]
x3 = mLib.tikhonov_regularization(A,b,[0.617])

r = np.sum(A*x[:,None,:],axis = -1)
r2= np.sum(A*x2[:,None,:],axis = -1)
r3= np.sum(A*x3[:,None,:],axis = -1)                     

print(np.mean(np.abs(r-b_free)/np.abs(b_free)))
print(np.mean(np.abs(r2-b_free)/np.abs(b_free)))
print(np.mean(np.abs(r3-b_free)/np.abs(b_free)))



import logging
import logging.handlers
import os
from xframe.settings import general

log_file_path = os.path.expanduser(general.log_file)
log_dir = os.path.dirname(log_file_path)
if not os.path.exists(log_dir):
    os.makedirs(log_dir)
existing_logs = {}

def setup_custom_logger(name,loglevelName):
    # logger settings
    if name in existing_logs:
        logger = existing_logs[name]
    else:
        log_file = log_file_path
        log_file_max_size = 1024 * 1024 * 20 # megabytes
        log_num_backups = 3
        log_format = "%(asctime)s [%(levelname)s]: %(filename)s(%(funcName)s:%(lineno)s) >> %(message)s"
        log_date_format = "%m/%d/%Y %I:%M:%S %p"
        log_filemode = "w" # w: overwrite; a:     
        loglevel=getattr(logging,loglevelName.upper())

        # setup logger
        # datefmt=log_date_format
        logging.basicConfig(filename=log_file, format=log_format, filemode=log_filemode ,level=loglevel)
        rotate_file = logging.handlers.RotatingFileHandler(
            log_file, maxBytes=log_file_max_size, backupCount=log_num_backups
        )
        logger = logging.getLogger(name)
        logger.propagate=False
        if len(logger.handlers)<3:
            logger.addHandler(rotate_file)
            #consoleHandler = logging.StreamHandler()
            # print log messages to console
            consoleHandler = logging.StreamHandler()
            logFormatter = logging.Formatter(log_format)
            consoleHandler.setFormatter(logFormatter)
            logger.addHandler(consoleHandler)
        logger.setLevel(loglevel)
        existing_logs[name]=logger        
    return logger

# source: https://docs.python.org/2/howto/logging.html



import numpy as np
import warnings
from scipy.stats import special_ortho_group
import math
from math import factorial
import logging
import inspect
import scipy.special as specialFunctions
from scipy.special import eval_jacobi
from scipy.special import hyp2f1
from scipy.special import roots_legendre
from scipy.special import comb as binomial
from scipy.linalg import polar as polar_decomp
from scipy.linalg import eigh
import sys
import abc
log=logging.getLogger('root')

from xframe.library.interfaces import GSLDependency
from xframe.library.interfaces import SphericalHarmonicTransformDependency
from xframe.library.interfaces import SoftDependency
from xframe.library.interfaces import DiscreteLegendreTransformDependency
from xframe.library.interfaces import PeakDetectorDependency
from scipy.stats import ttest_1samp
from scipy.optimize import root_scalar as sp_root

module_self = __import__(__name__)

gsl = GSLDependency
shtns = SphericalHarmonicTransformDependency
Soft = SoftDependency
leg_trf = DiscreteLegendreTransformDependency
nj = False
PeakDetector = PeakDetectorDependency


class plane3D:    
    base=np.empty(3)
    x_direction=np.empty(3)
    y_direction=np.empty(3)
    standardForm={'base':np.empty(3), 'x_direction':np.empty(3) , 'y_direction':np.empty(3)}

    def __init__(self,base=np.zeros(3),x_direction=np.array([1,0,0]),y_direction=np.array([0,1,0])):
        self._base=base
        self._x_direction=x_direction
        self._y_direction=y_direction
        self.standardForm='not set'
        self.updateStandardForm()
        self.normal=self._calc_normal()
    def _calc_normal(self):
        #selects normal pointing away from origin
        norm = np.linalg.norm
        n=np.cross(self._x_direction,self._y_direction)
        n_base_prod=n.dot(self.base)
        if not n_base_prod == 0:
            n*=n_base_prod/norm(n_base_prod)
            is_non_zero= np.nonzero(n)
            n[is_non_zero]/=norm(n[is_non_zero])
        return n
    
    def from_array(self,plane:np.ndarray):
        self.__init__(base=plane[0],x_direction=plane[1],y_direction=plane[2])
    
    def setPlain(self,base,x_direction,y_direction):
        self.base=base
        self.x_direction=x_direction
        self.y_direction=y_direction
        self.standardForm=self._calculateStandardForm()

    def updateStandardForm(self):
        normedY_direction=normalizeVector(self._y_direction)
        projectionOfXtoY=self._x_direction.dot(normedY_direction)
        standardX_direction=self._x_direction-projectionOfXtoY*normedY_direction
        standardX_direction=normalizeVector(standardX_direction)
        standardY_direction=normalizeVector(self._y_direction)
        self.standardForm={'base':self.base, 'x_direction':standardX_direction, 'y_direction':standardY_direction}

    @property
    def base(self):
        return self._base
    @base.setter
    def base(self,base:np.ndarray):
        self._base = base
        self.standardForm['base'] = self._base

    @property
    def x_direction(self):
        return self._x_direction
    @x_direction.setter
    def x_direction(self,x_direction:np.ndarray):
        self._x_direction = x_direction
        self.updateStandardForm()

    @property
    def y_direction(self):
        return self._y_direction
    @y_direction.setter
    def y_direction(self,y_direction:np.ndarray):
        self._y_direction = y_direction
        self.updateStandardForm()
    
        
class SampleShapeFunctions:
    coords_cartesian='cartesian'
    coords_polar='polar'
    coords_spherical='spherical'
    @classmethod
    def select_fct_by_coord_sys(cls,coord_sys,function_list):
        if coord_sys=='cartesian':
            #log.info('cartesian')
            function=function_list[0]
        elif coord_sys=='polar':
            #log.info('polar')
            function=function_list[1]
        elif coord_sys=='spherical':
            #log.info('spherical')
            function=function_list[2]
        else:
            log.error('Unknown coodinate System Type = {}.'.format(coordSys))
            raise AssertionError
        return function

    @classmethod
    def get_shape_function(cls,is_in_support,amplitude_function):
        def shape_function(points):
            #log.info('amplitude = {}'.format(amplitude_function(np.arange(100))))
            values=np.zeros(points[:].shape[:-1])
            support_mask=is_in_support(points)
            #log.info('support mask=\n{}'.format(support_mask))
            # log.info(points[support_mask].shape)
            # log.info(values[support_mask].shape)
            values[support_mask]=amplitude_function(points[support_mask])
            return values
        return shape_function

    @classmethod
    def get_disk_function(cls,radius,amplitude_function=lambda points: np.full(points.shape[:-1],1),coordSys='spherical',center=0,norm='standard',random_orientation=False):
        if norm=='standard':
            norm=np.linalg.norm
        elif norm == 'inf':
            def inf(points,axis=0):
                return np.max(np.abs(points),axis=axis)
            norm = inf
        so = special_ortho_group.rvs    
        def isInDisk_cartesian(points):
            if random_orientation:
                points = points[:].dot(so(points[:].shape[-1]))
#            log.info('point {} < radius {} = {}'.format(point,radius,point[0]<radius))
            return norm(points[:]-center,axis=-1)<radius
        def isInDisk_polar(points):
            cart_points=spherical_to_cartesian(points)[:]
            
            if isinstance(center,(np.ndarray,tuple,list)):
                cart_center=spherical_to_cartesian(np.array(center))
                log.info('cart center = {} input = {}'.format(cart_center,center))
            else:
                cart_center = 0
            cart_points-=cart_center
            if random_orientation:
                cart_points = cart_points.dot(so(cart_points.shape[-1]))
    
            #log.info('point {} < radius {} = {}'.format(point,radius,point[0]<radius))
            return norm(cart_points[:],axis=-1)<radius

        isInDisk=cls.select_fct_by_coord_sys(coordSys,[isInDisk_cartesian,isInDisk_polar,isInDisk_polar])
        disk_function=cls.get_shape_function(isInDisk,amplitude_function)
        return disk_function

    @classmethod
    def get_polygon_function(cls,radius,n_corners=3,amplitude_function=lambda points: np.full(points.shape[:-1],1),coordSys='polar',center = (0,0) , random_orientation=False):
        if random_orientation:
            orientation = np.random.rand()*2*np.pi
        else:
            orientation = 0
        corner_vectors=[]
        for n in range(n_corners):
            phi=(2*np.pi/n_corners*n+orientation)%(2*np.pi)
            corner_vector=radius*np.array([np.cos(phi),np.sin(phi)])+spherical_to_cartesian(np.array(center))
            corner_vectors.append(corner_vector)
            
        def isInPolygon_cartesian(points):
            is_in_polygon=True
            n_corners=len(corner_vectors)
            mask=True
            cart_center = spherical_to_cartesian(np.array(center))
            for n in range(n_corners):
                vec1=corner_vectors[n]
                vec2=corner_vectors[(n+1)%n_corners]
                #vec1-=cart_center
                #log.info('center = {}vec 1 = {}'.format(cart_center,vec1))
                #vec2-=cart_center
                #log.info(points.shape)
                #points-=np.array(cart_center)[None,None,:]
                #rotate90_matrix=np.array([[0,1],[-1,0]])
                rotate90_matrix=np.array([[0,1],[-1,0]])
                diff=vec2-vec1
                scal_prod=np.tensordot((points-vec1),rotate90_matrix.dot(diff),axes=([-1],[0]))
                #scal_prod=np.tensordot((points-vec1),diff,axes=([-1],[0]))                
                mask*=np.where(scal_prod>0,False,True)
            return mask
        def isInPolygon_polar(points):
            cart_points = spherical_to_cartesian(points)
            mask=isInPolygon_cartesian(cart_points)
#            log.info('Point {} is in polygon: {}'.format([r*np.cos(phi),r*np.sin(phi)],is_in_polygon))
            return mask
        
        isInPolygon=cls.select_fct_by_coord_sys(coordSys,[isInPolygon_cartesian,isInPolygon_polar,None])
        polygon_function=cls.get_shape_function(isInPolygon,amplitude_function)
        return polygon_function
        
    @classmethod
    def get_tetrahedral_function(cls,radius,amplitude_function=lambda points: np.full(points.shape[:-1],1),coordSys='spherical',center=np.zeros(3), orientation=np.eye(3),random_orientation=False):        
        if random_orientation:
            orientation = special_ortho_group.rvs(3)   
        planes = cls.get_tetrahedron_planes(radius = radius)
        is_in_thetrahedron_bare = cls.get_is_in_shape_from_planes(planes,orientation=orientation)
        def is_in_thetrahedron_cart(points):
            points -= center
            shape = is_in_thetrahedron_bare(points)
            return shape
        
        def is_in_thetrahedron_spherical(points):
            cart_points = spherical_to_cartesian(points)[:]
            cart_center = spherical_to_cartesian(np.asarray(center))
            cart_points -= cart_center 
            shape = is_in_thetrahedron_bare(cart_points)
            #log.info('point {} < radius {} = {}'.format(point,radius,point[0]<radius))
            return shape

        is_in_tetrahedron=cls.select_fct_by_coord_sys(coordSys,[is_in_thetrahedron_cart,None,is_in_thetrahedron_spherical])
        tetrahedral_function=cls.get_shape_function(is_in_tetrahedron,amplitude_function)
        return tetrahedral_function

    @classmethod
    def get_is_in_shape_from_planes(cls,planes,orientation = np.eye(3)):
        SO = special_ortho_group.rvs
        def is_in_shape(cart_points):
            if not isinstance(orientation,np.ndarray):
                rot = SO(3)
            else:
                rot = orientation
            cart_points = cart_points.dot(rot)            
            shape = np.array([True])
            for plane in planes:
                #orth vector from plane to point in units of normal of plane
                d=plane.base.dot(plane.normal)-cart_points[:].dot(plane.normal)
                in_cut = (d>=0)
                shape = (shape & in_cut)
            return shape
        return is_in_shape
    
    @classmethod
    def get_tetrahedron_planes(cls,radius=1):
        z=1/np.sqrt(2)
        R=radius/np.linalg.norm(np.array([0,-1,-z]))
        b1 = R*np.array([0,-1,-z])
        b2 = R*np.array([-1,0,z])
        x1 = np.array([0,2.,0])
        x2 = np.array([2.,0,0])
        y1 = (b1/R-np.array([1.,0,z]))
        y2 = (b1/R-np.array([-1.,0,z]))
        y3 = (b2/R-np.array([0,-1,-z]))
        y4 = (b2/R-np.array([0,1,-z]))
        
        p1=plane3D(base=b1,x_direction = x1,y_direction = y1)
        p2=plane3D(base=b1,x_direction = x1,y_direction = y2)
        p3=plane3D(base=b2,x_direction = x2,y_direction = y3)
        p4=plane3D(base=b2,x_direction = x2,y_direction = y4)
        return [p1,p2,p3,p4]

    @classmethod
    def get_rectangle_function(cls,lengths,center = np.zeros(3),amplitude_function=lambda points: np.full(points.shape[:-1],1),coordSys='spherical'):        
        def is_in_rectangle_cartesian(points):
            n_dims = len(lengths)
            limits = tuple([center[i]-lengths[i]/2,center[i]+lengths[i]/2] for i in range(n_dims))
            mask = True
            for i in range(n_dims):
                p = points[...,i]
                limit = limits[i]
                mask &= ((p>limit[0]) & (p<limit[1]))
            #log.info('rectangle_mask shape = {}'.format(mask.shape))        
            return mask
        def is_in_rectangle_polar(points):
            cart_points = spherical_to_cartesian(points)
            cart_center = spherical_to_cartesian(np.array(center))
            n_dims = len(lengths)
            limits = tuple([cart_center[i]-lengths[i]/2,cart_center[i]+lengths[i]/2] for i in range(n_dims))
            mask = True
            for i in range(n_dims):
                p = cart_points[...,i]
                limit = limits[i]
                mask &= ((p>limit[0]) & (p<limit[1]))
            return mask
        is_in_rectangle=cls.select_fct_by_coord_sys(coordSys,[is_in_rectangle_cartesian,is_in_rectangle_polar,is_in_rectangle_polar])
        rectangle_function=cls.get_shape_function(is_in_rectangle,amplitude_function)
        return rectangle_function
    @classmethod
    def get_anulus_function(cls,inner_radius,outer_radius,center = np.zeros(2),amplitude_function=lambda points: np.full(points.shape[:-1],1),coordSys='spherical',norm = 'standard'):
        if inner_radius == 0:
            anulus_function = cls.get_disk_function(outer_radius,amplitude_function = amplitude_function,coordSys = coordSys, center = center,norm = norm)
        else:
            if norm=='standard':
                norm=np.linalg.norm
            elif norm == 'inf':
                def norm(points,axis=0):
                    return np.max(np.abs(points),axis=axis)
                
            def is_in_anulus_cartesian(points):
                length=norm(points[...,:]-center,axis=-1)
                return (length>inner_radius) & (length<outer_radius)
            def is_in_anulus_polar(points):
                cart_points=spherical_to_cartesian(points)[...,:]
                cart_center=spherical_to_cartesian(np.array(center))
                length = norm(cart_points[:] - cart_center,axis=-1)
                return (length>inner_radius) & (length<outer_radius)
            is_in_anulus=cls.select_fct_by_coord_sys(coordSys,[is_in_anulus_cartesian,is_in_anulus_polar,is_in_anulus_polar])
            anulus_function=cls.get_shape_function(is_in_anulus,amplitude_function)
        return anulus_function

    
def output_lowValuesToZero_decorator(function):
    def lowValuesToZero(array):
        tolerance=1e-15
        isComplex=array.dtype==np.complex
        if isComplex:
            array.real[np.abs(array.real)<tolerance]=0.
            array.imag[np.abs(array.imag)<tolerance]=0.
        else:
            array[np.abs(array)<tolerance]=0.
        return array    
    def newFunction(*args,**kwargs):
        output=function(*args,**kwargs)
        output=lowValuesToZero(output)
        return output

    #applies the parameter signature of function to newFunction
    undecoratedParameters=inspect.signature(function).parameters.values()
    newFunction.__signature__=inspect.signature(newFunction).replace(parameters=undecoratedParameters)
    return newFunction

@output_lowValuesToZero_decorator
def fft_complex_mixedRadix_forward(complex_data_array, step_length=1):    
    transformed_data=np.fft.fft(complex_data_array)
#    log.info('data mathlib forward out =\n{}'.format(newData))
    return transformed_data*step_length

@output_lowValuesToZero_decorator
def fft_complex_mixedRadix_inverse(complex_data_array, step_length=1):
    transformed_data=np.fft.ifft(complex_data_array)
#    log.info('data mathlib inverse out =\n{}'.format(transformedData))
    return transformedData*step_length

def fft_complex_radix2_inverse(complexDataArray, stepLength):
    numberOfSamples=complexDataArray.size/2
    tempData=np.copy(complexDataArray)/stepLength
    gsl.gsl_fft_complex_radix2_inverse(ctypes.c_void_p(tempData.ctypes.data), np.int(1), numberOfSamples)
    tempData=tempData
    return tempData

def fft_real_radix2(realDataArray, stepLength):
    numberOfSamples=realDataArray.size
    tempData=np.copy(realDataArray)
    gsl.gsl_fft_real_radix2_transform(ctypes.c_void_p(tempData.ctypes.data), np.int(1), numberOfSamples)
    tempData=tempData*stepLength
    return tempData


@output_lowValuesToZero_decorator
def fft_real_mixedRadix_forward(realDataArray,stepLength=1):
    newData=np.copy(realDataArray).astype(np.float)
    transformedData=gsl.fft_real_mixedRadix_forward(newData)
    return transformedData*stepLength

@output_lowValuesToZero_decorator
def fft_halfcomplex_mixedRadix_inverse(dataArray,stepLength=1):
    newData=np.copy(dataArray).astype(np.float)
    transformedData=gsl.fft_halfcomplex_mixedRadix_inverse(newData)
    return transformedData*stepLength


def fft_halfcomplex_unpack(halfcomplexPackedArray):
    unpackedArray=gsl.fft_halfcomplex_unpack(halfcomplexPackedArray)
    return unpackedArray


def fft_halfcomplex_pack(complexArray):
    lenData=len(complexArray)   
    packedArray=np.dstack((complexArray.real,complexArray.imag)).flatten()
    lenDataIsEven= lenData%2==0
    if lenDataIsEven:        
        halfcomplexArray=np.concatenate((np.array([packedArray[0]]),packedArray[2:lenData+1]))
#        log.info('halfcomplexArray={}'.format(halfcomplexArray))
    else:
        halfcomplexArray=np.concatenate((packedArray[0],packedArray[2:lenData+1]))
    return halfcomplexArray



def generate_dht_2d(maxR,n_radial_points,harmonic_orders):
    dht,inverse_dht=gsl.generate_dht_2D(n_radial_points,harmonic_orders,maxR)
    return dht,inverse_dht

@output_lowValuesToZero_decorator
def dht_forward(complexDataArray,nu):
    transformedArray=gsl.dht_forward(complexDataArray,nu)
    return transformedArray

@output_lowValuesToZero_decorator
def dht_inverse(complexDataArray,nu):
    transformedArray=gsl.dht_inverse(complexDataArray,nu)
    return transformedArray

def dht_bare(complexDataArray,nu,maxR=1):
    transformedArray=gsl.dht_bare(complexDataArray,nu,maxR=maxR)
    return transformedArray


def besselZeros(orders,zeros):
    ordersArray,zerosArray=np.meshgrid(orders,zeros,indexing='ij')
    ordersArray=np.abs(ordersArray)
    zeros=np.array(list(map(gsl.besselZero,ordersArray.flatten(),zerosArray.flatten()))).reshape(len(orders),len(zeros))
    return zeros
def besselZeros2(order,numberOfZeros):
    zeros=specialFunctions.jn_zeros(order,numberOfZeros)
    return zeros

def bessel_jnu(*args,**kwargs):
    value=specialFunctions.jv(*args,**kwargs)
    return value

def spherical_bessel_jnu(*args,**kwargs):
    value=specialFunctions.spherical_jn(*args,**kwargs)
    return value

def eval_legendre(degrees,arguments):
#    value=gsl.legendrePoly(degree,x)
    return specialFunctions.eval_legendre(degrees,arguments)

def legendrePlFunction(degree):
    polynomial=specialFunctions.legendre(degree)
    return polynomial

@output_lowValuesToZero_decorator
def fft_realToComplex_mixedRadix_forward(realDataArray):
    data=np.copy(realDataArray)
    gsl.fft_real_mixedRadix_forward(data)
    complexData=gsl.fft_halfcomplex_unpack(data)
    return complexData

@output_lowValuesToZero_decorator
def fft_complexToReal_mixedRadix_inverse(complexDataArray):
    packedData=fft_halfcomplex_pack(complexDataArray)
    gsl.fft_halfcomplex_mixedRadix_inverse(packedData)
    realData=packedData
    return realData


def circularHarmonicTransform_real_forward_gsl(dataArray):
    data=np.copy(dataArray)
    harmonicCoefficients=gsl.fft_real_mixedRadix_forward(data)/len(data)    
    return harmonicCoefficients

def circularHarmonicTransform_halfcomplex_inverse_gsl(dataArray,even=True):
    data=np.copy(dataArray)
    realData=gsl.fft_halfcomplex_mixedRadix_inverse(data)/len(data)
    return realData


def circularHarmonicTransform_complex_forward(data_array):
    data=np.copy(data_array)
    #   log.info('circular harmonic forward data shape={}'.format(data.shape))
    harmonic_coefficients=np.fft.fft(data,axis = 1)/data.shape[-1]
    #    harmonicCoefficients=np.fft.fft(data)/len(data)

    return harmonic_coefficients


def circularHarmonicTransform_complex_inverse(data_array):
    data=np.copy(data_array)*data_array.shape[-1]
    real_data=np.fft.ifft(data,axis = 1)
#    realData=np.fft.ifft(data)
    return real_data

def circularHarmonicTransform_real_forward(data_array):
    #log.info("sum input = {}".format(np.sum(data_array)))
    data=np.copy(data_array.real)
    #    log.info('circular harmonic forward data shape={}'.format(data.shape))
    harmonic_coefficients=np.fft.rfft(data)/data.shape[-1]
    #log.info('data shape = {}'.format(data.shape[-1]))
    return harmonic_coefficients

def circularHarmonicTransform_real_inverse(data_array,size):
    data=np.copy(data_array)*size
    #log.info('data shape = {}'.format(size))
    real_data=np.fft.irfft(data,size)
    return real_data

def get_spherical_harmonic_transform_obj(l_max,mode='complex',anti_aliazing_degree=2,n_phi=False,n_theta = False):
    sh=shtns(l_max,mode_flag=mode,anti_aliazing_degree = anti_aliazing_degree,n_phi=n_phi,n_theta=n_theta)
    return sh

def get_soft_obj(l_max):
    global Soft
    soft=Soft(l_max)
    return soft

def complexArraytoRealArray(array):
    realRepresentation=np.stack((array.real,array.imag)).flatten()
    return realRepresentation

def normalizeVector(vector):
    try:
        normedVector=vector*1/np.linalg.norm(vector)
    except ZeroDivisionError as e:
        print('Zero vector can not be normalized')
        print(e)
    return normedVector

def isAlmostZero(number):
    if number < 10**(-15):
        return True
    else:
        return False


def gauss_legendre(n,start=-1,stop=1):
    '''
    returns gauss legendre nodes and weights for integration range from start to stop.
    '''
    xi,w = roots_legendre(n)
    xi = (stop-start)/2*xi+(start + stop)/2
    w = (stop-start)/2*w
    return xi,w
    

def gaussian_fouriertransformed_polar2D_array(pointsArray,sigma):
    valueArray=np.exp(-2*(np.pi**2)*(sigma**2)*(np.square(pointsArray[...,0])) )
    return valueArray
def gaussian_cart(points,sigma):
    '''
    Evaluates a gaussian centered at 0 in cartesian coordinates.
    '''
    if points.ndim ==1:
        points=points[:,None]
    a=1/(2*sigma**2)
    values = np.exp(-np.linalg.norm(points,axis = -1)**2*a)
    return values
def gaussian_spher(points,sigma):
    '''
    Evaluates a gaussian centered at 0 in spherical coordinates.
    '''
    a=1/(2*sigma**2)
    values = np.exp(-np.square(points[...,0])**2*a)
    return values
def gaussian_fourier_transformed_cart(points,sigma):
    '''
    Evaluates the fourier transform of a gaussian centered at 0 in cartesian coordinates.
    '''
    if points.ndim ==1:
        points=points[:,None]
    dim = points.shape[-1]
    pi = np.pi
    a=1/(2*sigma**2)
    prefactor = np.sqrt(pi/a)*1/(sigma*np.sqrt(2*pi))
    values = prefactor**dim * np.exp(-np.pi**2 * np.linalg.norm(points,axis = -1)**2 / a)
    return values
def gaussian_fft_cart(points,sigma):
    if points.ndim ==1:
        points=points[:,None]
    sizes = get_fft_grid_sizes(points)
    values = gaussian_fourier_transformed_cart(points,sigma)
    log.info('points.shape = {}'.format(points.shape))
    phases = 1
    #phases = np.prod((-1)**np.array(np.meshgrid(*[np.concatenate((np.arange(n//2 +1),np.arange(1,n//2+n%2)[::-1])) for n in points.shape[:-1]],indexing = 'ij')[::-1]).T,axis = -1)
    #phases = 1#np.prod((-1)**np.array(np.meshgrid(*[np.arange(n) for n in points.shape[:-1]],indexing = 'ij')[::-1]).T,axis = -1)
    phases = np.squeeze(phases)
    log.info('phases.shape = {}'.format(phases.shape))
    log.info('sizes = {}'.format(sizes))
    values*= phases * np.prod(sizes)
    return values
def get_fft_grid_steps(points):
    if points.ndim ==1:
        points=points[:,None]
    dim = points.shape[-1] 
    steps = np.abs(np.diag(points[tuple(np.eye(dim,dtype = int))]))
    return steps
def get_fft_grid_sizes(points):
    if points.ndim ==1:
        points=points[:,None]
    steps = get_fft_grid_steps(points)
    sizes = steps * points.shape[:-1]
    return sizes
def get_fft_reciprocal_grid(grid):
    if grid.ndim ==1:
        grid=grid[:,None]
    sizes = get_fft_grid_sizes(grid)
    length = grid.shape[:-1]
    new_sizes = 2/(2*sizes)
    
    reciprocal_grid = np.array(
        np.meshgrid(*[np.concatenate((np.arange(n//2 +1),-np.arange(1,n//2+n%2)[::-1])) for n in length],indexing = 'ij')[::-1]
        ,dtype = float).T
    reciprocal_grid *= new_sizes
    return np.squeeze(reciprocal_grid)

def convolve_with_gaussian_cart(data,grid,sigma=1):
    r_grid = get_fft_reciprocal_grid(grid)
    r_data = np.fft.fftn(data)
    r_gaussian = gaussian_fft_cart(r_grid,sigma)
    convolution = np.fft.ifftn(r_data*r_gaussian).real
    data_sum = np.sum(data)
    convolution_sum = np.sum(convolution)
    convolution *= data_sum/convolution_sum
    return convolution

def gaussian_fourier_transformed_spherical(points,sigma):
    '''
    Evaluates the fourier transform of a gaussian centered at 0 in spherical coordinates.
    '''
    pi = np.pi
    a=1/(2*sigma**2)
    prefactor = np.sqrt(pi/a) 
    values = prefactor * np.exp(-np.pi**2 * np.square(points[...,0])**2 / a)
    return values

def id(_in):
    return _in

def cartesian_to_spherical(grid):
    dimension=grid[:].shape[-1]
    n_grid=grid.copy()
    try:
        assert dimension in (2,3) 
    except AssertionError as e:
        log.error(e)
        raise e

    if dimension==2:
        x=grid[...,0].copy()
        y=grid[...,1].copy()

        n_grid[...,0]=np.sqrt(np.square(x)+np.square(y))
        phi=np.arctan2(y,x)
        n_grid[...,1]=np.where(phi<0,phi+2*np.pi,phi)        
    else:
        x=grid[...,0].copy()
        y=grid[...,1].copy()
        z=grid[...,2].copy()

        r=np.sqrt(np.square(x)+np.square(y)+np.square(z))
        n_grid[...,0]=r
        theta=np.zeros(grid[...,0].shape)
        r_nonzero=(r!=0)
        if r_nonzero.any():
            #log.info('r_nonzero={}'.format(r_nonzero))
            tmp=z[r_nonzero]/r[r_nonzero]
            #log.info('shape tmp = {}'.format(tmp.shape))
            #log.info('phi non_zero shape = {}'.format(phi[r_nonzero].shape))
            theta[r_nonzero]=np.arccos(tmp)
        #log.info('phi={}'.format(theta))
        #log.info('n_grid={}'.format(n_grid))
        n_grid[...,1]=theta
        phi=np.arctan2(y,x)
        n_grid[...,2]=np.where(phi<0,phi+2*np.pi,phi)
    return n_grid

def get_gaussian_weights_1d(data,grid_points):
    sum_data=np.sum(data)
    pos=np.sum(data*grid_points)/sum_data
    sigma=np.sqrt(np.abs(np.sum(np.square(grid_points-pos)*data)/sum_data))
    return {'pos':pos,'sigma':sigma}

def spherical_to_cartesian(grid):
    dimension=grid[:].shape[-1]
    n_grid=grid.copy()
    try:
        assert dimension in (2,3) 
    except AssertionError as e:
        log.error(e)
        raise e

    if dimension==2:
        r=grid[...,0].copy()
        phi=grid[...,1].copy()

        n_grid[...,0]=r*np.cos(phi)
        n_grid[...,1]=r*np.sin(phi)
    else:
        r=grid[...,0].copy()
        theta=grid[...,1].copy()
        phi=grid[...,2].copy()

        xy_projection=r*np.sin(theta)

        n_grid[...,0]=np.cos(phi)*xy_projection
        n_grid[...,1]=np.sin(phi)*xy_projection
        n_grid[...,2]=r*np.cos(theta)
    return n_grid
        
def cylindricalToCartesian(Data):
    x=np.cos(Data[:,1])*Data[:,0]
    y=np.sin(Data[:,1])*Data[:,0]
    z=Data[:,2]
    cartesianData=np.stack((x,y,z),axis=1)
    return cartesianData
 
def cartesianToSpherical(Data):
     r=np.linalg.norm(Data,axis=1)
     
     nonZeroRadi_mask= r>0
     theta=np.zeros(r.shape)
     theta[nonZeroRadi_mask]=np.arccos(Data[:,2][nonZeroRadi_mask]/r[nonZeroRadi_mask])

     xPositions=Data[:,0]
     yPositions=Data[:,1]
     positiveY_mask= yPositions>=0
     negativeY_mask= np.logical_not(positiveY_mask)
     thetaNot0orPi_mask=np.logical_and(theta!=0,theta!=np.pi)

     positiveY_NonZeroDiv_mask=np.logical_and(np.logical_and(positiveY_mask,nonZeroRadi_mask),thetaNot0orPi_mask)

     negativeY_NonZeroDiv_mask=np.logical_and(np.logical_and(negativeY_mask,nonZeroRadi_mask),thetaNot0orPi_mask)
     phi=np.zeros(r.shape)
     phi[positiveY_NonZeroDiv_mask]=np.arccos(xPositions[positiveY_NonZeroDiv_mask]/r[positiveY_NonZeroDiv_mask])
     phi[negativeY_NonZeroDiv_mask]=2*np.pi-np.arccos(xPositions[negativeY_NonZeroDiv_mask]/r[negativeY_NonZeroDiv_mask])

     sphericalData=np.stack((r,theta,phi),axis=1)
     return sphericalData

def phiFrom_R_Theta_X_Y(r,theta,x,y):
     nonZeroRadi_mask= r>0
     positiveY_mask= y>=0
     negativeY_mask= np.logical_not(positiveY_mask)
     thetaNot0orPi_mask=np.logical_and(theta!=0,theta!=np.pi)
     positiveY_NonZeroDiv_mask=np.logical_and(np.logical_and(positiveY_mask,nonZeroRadi_mask),thetaNot0orPi_mask)
     negativeY_NonZeroDiv_mask=np.logical_and(np.logical_and(negativeY_mask,nonZeroRadi_mask),thetaNot0orPi_mask)
     
     phi=np.zeros(r.shape)
     phi[positiveY_NonZeroDiv_mask]=np.arccos(x[positiveY_NonZeroDiv_mask]/r[positiveY_NonZeroDiv_mask])
     phi[negativeY_NonZeroDiv_mask]=2*np.pi-np.arccos(x[negativeY_NonZeroDiv_mask]/r[negativeY_NonZeroDiv_mask])
     return phi

def testAdd(a,b):
    return a+b


def second_derivative(data,step_size,accuracy_order):
    '''
    Approximates the second derivative for a dataset sampled on a uniform grid for accuracy_orders in (2,4,6,8):
    See https://en.wikipedia.org/wiki/Finite_difference_coefficient
    '''
    coefficients_by_order = {
        '2':np.array([1,-2,1]),
        '4':np.array([-1/12,4/3,-5/2,4/3,-1/12]),
        '6':np.array([1/90,-3/20,3/2,-49/18,3/2,-3/20,1/90]),
        '8':np.array([-1/560,8/315,-1/5,8/5,-205/72,8/5,-1/5,8/315,-1/560])
    }
    
    assert str(accuracy_order) in coefficients_by_order, 'Invalid accuracy_order {} valid inputs are {}'.format(accuracy_order,coefficients_by_order.keys())
    return np.convolve(data,coefficients_by_order[str(accuracy_order)],mode='valid')/step_size**2

def first_derivative(data,step_size,accuracy_order):
    '''
    Approximates the second derivative for a dataset sampled on a uniform grid for accuracy_orders in (2,4,6,8):
    See https://en.wikipedia.org/wiki/Finite_difference_coefficient
    '''
    coefficients_by_order = {
        '2':np.array([-1/2,0,1/2]),
        '4':np.array([1/12,-2/3,0,2/3,-1/12]),
        '6':np.array([-1/60,3/20,-3/4,0,3/4,-3/20,1/60]),
        '8':np.array([1/280,-4/105,1/5,-4/5,0,4/5,-1/5,4/105,-1/280])
    }
    
    assert str(accuracy_order) in coefficients_by_order, 'Invalid accuracy_order {} valid inputs are {}'.format(accuracy_order,coefficients_by_order.keys())
    return -np.convolve(data,coefficients_by_order[str(accuracy_order)],mode='valid')/step_size
    
    

def eval_zernike_polynomials(m_array,n_max,points):
    def evaluate(m,n):
        return ((-1)**((n-m)/2))*(points**m)*eval_jacobi((n-m)/2,m,0,1-2*(points**2))
#        return (points**m)#*eval_jacobi((n-m)/2,m+1,m+1,points**2)/eval_jacobi((n-m)/2,m+1,m+1,1)
    values={}
    for m in m_array:
        print('generating zernike polynomials for m={}'.format(m))
        ns=np.arange(m,n_max+1,2)
        len_ns=len(ns)
        values[m]=np.array(tuple(map(evaluate,(m,)*len_ns,ns)))
    return values


def eval_ND_zernike_polynomials_old(m_array,n_max,points,dimension):
    D=dimension
    def evaluate(m,n):
        return ((-1)**((n-m)/2))*(points**m)*eval_jacobi((n-m)/2,m+D/2-1,0,1-2*(points**2))
#        return (points**m)#*eval_jacobi((n-m)/2,m+1,m+1,points**2)/eval_jacobi((n-m)/2,m+1,m+1,1)
    values={}
    for m in m_array:
        #print('generating zernike polynomials for m={}'.format(m))
        ns=np.arange(m,n_max+1,2)
        len_ns=len(ns)
        values[m]=np.array(tuple(map(evaluate,(m,)*len_ns,ns)))
    return values

def eval_ND_zernike_polynomials(l_array,s_max,points,dimension):
    '''
    Computes the radial part $R^l_s(p)$ of the D-dimensional Zernike polynomials. For each $l$ in l\_ array all coefficients $R^l_s(p)$ with $s_{\max} \geq s\geq l$ and even $s-l$ are computed.    
    '''
    D=dimension
    def evaluate(l,s):
        return ((-1)**((s-l)/2))*(points**l)*eval_jacobi((s-l)/2,l+D/2-1,0,1-2*(points**2))
#        return (points**m)#*eval_jacobi((n-m)/2,m+1,m+1,points**2)/eval_jacobi((n-m)/2,m+1,m+1,1)
    values={}
    for l in l_array:
        #print('generating zernike polynomials for m={}'.format(m))
        s=np.arange(l,s_max+1,2)
        len_s=len(s)
        values[l]=np.array(tuple(map(evaluate,(l,)*len_s,s)))
    return values

def eval_ND_zernike_polynomials_new(m_array,n_max,points,dimension):
    D=dimension
    def evaluate(m,n):
        nmD2=(n+m+D)/2
        nm2 = (n-m)/2
        binom_args=[nmD2-1,nm2]
        f1_args = [-nm2,nmD2,m+D/2,points**2]
        return (-1)**nm2*(points**m)*hyp2f1(*f1_args)*binomial(*binom_args)
        #return ((-1)**((n-m)/2))*(points**m)*eval_jacobi((n-m)/2,m-1+D/2,0,1-2*(points**2))
#        return (points**m)#*eval_jacobi((n-m)/2,m+1,m+1,points**2)/eval_jacobi((n-m)/2,m+1,m+1,1)
    values={}
    for m in m_array:
        print('generating zernike polynomials for m={}'.format(m))
        ns=np.arange(m,n_max+1,2)
        len_ns=len(ns)
        values[m]=np.array(tuple(map(evaluate,(m,)*len_ns,ns)))
    return values

def eval_ND_zernike_polynomials_gsl(m_array,n_max,points,dimension):
    D=dimension
    def evaluate(m,n):
        nmD2=(n+m+D)/2
        nm2 = (n-m)/2
        binom_args=[nmD2-1,nm2]
        f1_args = [-nm2,nmD2,m+D/2,points**2]
        h2f1 = gsl.hyperg_2F1(*f1_args)
        return (-1)**nm2*(points**m)*h2f1*binomial(*binom_args)
        #return ((-1)**((n-m)/2))*(points**m)*eval_jacobi((n-m)/2,m-1+D/2,0,1-2*(points**2))
#        return (points**m)#*eval_jacobi((n-m)/2,m+1,m+1,points**2)/eval_jacobi((n-m)/2,m+1,m+1,1)
    values={}
    for m in m_array:
        print('generating zernike polynomials for m={}'.format(m))
        ns=np.arange(m,n_max+1,2)
        len_ns=len(ns)
        values[m]=np.array(tuple(map(evaluate,(m,)*len_ns,ns)))
    return values



def generate_diagonal_circular_ht(n_angular_points,n_orders=False):
    if isinstance(n_orders,bool):
        n_orders=n_angular_points
    pt=np.arange(n_angular_points)
    o=np.concatenate((np.arange(n_orders//2+1),-1*np.arange(n_orders//2+n_orders%2)[:0:-1]))
    exp=1/n_angular_points*np.exp(-1.j*2*np.pi*o[:,None]*pt[None,:]/n_angular_points)

    def cht(f):
        f=f.reshape(n_orders,-1,n_angular_points)
        return np.sum(f*exp[:,None,:],axis=2)
    return cht

def nearest_positive_semidefinite_matrix(A,low_positive_eigenvalues_to_zero = False):
    '''
    Generates the closest matrix to A in the Frobenius norm that is positive semidefinite.
    See: N.J. Higham, "Computing a nearest symmetric positive semidefinite
    matrix" (1988): https://doi.org/10.1016/0024-3795(88)90223-6 
    '''
    B = (A + np.swapaxes(A,-1,-2).conj()) / 2
    l,v = np.linalg.eigh(B)#,driver = 'ev')
    #l,v = np.linalg.eigh(B)
    limit = 0
    if low_positive_eigenvalues_to_zero:
        # Estimates eigenvalue noise floor by considering the absolute value of the lowest eigenvalue of A
        # project all eigenvalues with lower values to zero.
        eigvals,_ = np.linalg.eig(A)
        limit = np.abs(np.min(eigvals))
        #limit = 0
    #log.info(f'eig limit = {limit}')
    l[l<limit]=0
    A2 =v*l[...,None,:] @ np.swapaxes(v,-1,-2).conj()
    #A2 = v @ np.diag(l) @ v.T
    return A2

def approximate_tikhonov_parameters(A,b,std_noise=False):
    '''
    Chooses optimal tikhonov parameters as described in
    [1] O'leary, Dianne. (2001). Near-Optimal Parameters for Tikhonov and Other Regularization Methods. SIAM Journal on Scientific Computing. 23. 10.1137/S1064827599354147. 
    '''
    if A.ndim < 3:
        A = A[None,...]
        b = b[None,...]
    m,n = A[0].shape
    if isinstance(std_noise,bool):
        size_for_std = max(m-n,10)
        std_estimate = np.std(b[:,-size_for_std:],axis = -1)
    else:
        std_estimate = np.atleast_1d(std_noise)
    u,s,vh = np.linalg.svd(A,full_matrices=False)
    betas = np.sum(u*b[:,:,None],axis = -2)
    ks = []
    for i,beta in enumerate(betas):
        if abs(b[i,n-1]) > 3.5*std_estimate[i]:
            ks.append(n)
        else:
            k=n
            for j in np.arange(n-1)[::-1]:
                p_value = ttest_1samp(beta[j:n].real,0)[1]
                if p_value>0.05:
                    k=j
            ks.append(k)
    ks = np.asarray(ks)
    def parameter_func(lambd,i):
        D = (s[i]**2+lambd)
        #log.info('A shape = {} betas shape = {} std_estimate_shape ={} ks shape = {}'.format(s.shape,betas.shape,std_estimate.shape,ks.shape))
        val = np.sum(np.abs(betas[i])**2*lambd/(D**3)) - np.sum((np.abs(betas[i])**2/(D**2))[ks[i]:]) - np.sum((std_estimate[i]**2/(D**2))[:ks[i]])
        #print(val)
        return val
    
    def find_root_bracket(i):
        std = std_estimate[i]
        lower_bound = 0
        upper_bound = False       
        for h in np.arange(30)[::-1]:
            lower_lambd = std/(10**h)
            upper_lambd = std*(10**h)
            x = parameter_func(lower_lambd,i)
            y = parameter_func(upper_lambd,i)
            if x<0:
                lower_bound =lower_lambd
            if y>=0:
                upper_bound = upper_lambd
            #print("b = {} x = {},b = {} y = {}".format(lower_lambd,x,upper_lambd,y))
        #log.info('lower {}, upper {}, last x {},last y {}'.format(lower_bound,upper_bound,x,y))
        try:
            assert isinstance(upper_bound,float),'could not find upper bound on the tikhonov parameter for the {}th linear system,'.format(i)
        except Exception as e:
            log.warning('Could not find optimal tikhonov parameter for the {}th linear system. Default to lambda = 0.0.last tested value = {}'.format(i,y))
            return False
        return (lower_bound,upper_bound)
    def find_root(system_id):
        bracket = find_root_bracket(system_id)
        if isinstance(bracket,tuple):
            lambd = sp_root(parameter_func,args = (system_id,),bracket = bracket).root            
        else:
            lambd = 0.0
        return lambd
    lambd = tuple(find_root(i) for i in range(len(A)))
    #log.info(lambd)
    return np.asarray(lambd)
    
def tikhonov_solver_svd(A,b,lambd):
    '''
    Finds x that minimizes $|| A x -B ||^2 - ||\lambda Id x ||^2 $
    '''
    n_orders = A.shape[-1]
    lambd = np.atleast_1d(lambd)
    if A.ndim < 3:
        A = A[None,...]
        b = b[None,...]

    u, s, vh = np.linalg.svd(A, full_matrices=False)
    small_singular_values = s <= 1e-15  # same default value as scipy.linalg.pinv
    uTb = np.sum(u*b[:,:,None],axis = -2)
    d = s / (s**2 + lambd[:,None])
    d[small_singular_values] = 0
    d_uTb = d * uTb
    x = np.sum(vh*d_uTb[:,:,None],axis = -2)        
    return x

def tikhonov_prepare_data(A,b,weights = None):
    A = np.copy(A)
    b = np.copy(b)

    A_offsets = np.average(A, axis=1,weights=weights)
    A -= A_offsets[:,None,:]    
    b_offsets = np.average(b, axis=1,weights=weights)
    b -= b_offsets[:,None]
    return A,b,A_offsets,b_offsets

def tikhonov_calc_offset(xs,A_offsets,b_offsets):
    return b_offsets - np.sum(A_offsets*xs,axis = -1)    

def tikhonov_regularization(As,bs,lambd,allow_offset = False):
    '''  
    For each pair of elements (A,b) in As,bs this method finds x that minimizes: 
    $||Ax-b||+||\lambda*x||$    
    if allow_offset is true it minimizes:
    $||A'x-b'||+||\lambda*x||$ 
    where A'=A-np.mean(A,axes = 0) and b' = b-np.mean(b).
    For that x one has $Ax+offset \approx b$, where the offset is fixed by x,A-A'and b-b'.    
    '''
    if allow_offset:
        As,bs,A_offsets,b_offsets = tikhonov_prepare_data(As,bs)
        xs = tikhonov_solver_svd(As,bs,lambd)        
        offsets = tikhonov_calc_offset(xs,A_offsets,b_offsets)
    else:
        xs = tikhonov_solver_svd(As,bs,lambd)
        offsets = np.zeros(len(As),dtype = As.dtype)
    return xs,offsets    

def optimal_tikhonov_regularization(As,bs,allow_offset = False,std_noise = False):    
    lambdas = approximate_tikhonov_parameters(As,bs,std_noise=std_noise)
    xs,offsets = tikhonov_regularization(As,bs,lambdas,allow_offset=allow_offset)
    return xs,offsets
def optimal_tikhonov_regularization_worker(q1s,q2s,As,bs,allow_offset = False,std_noise = False,**kwargs):
    #log.info(q1s)
    #log.info(q2s)
    As = As[q1s,q2s]
    bs = bs[q1s,q2s]
    #log.info('As shape = {} bs shape = {}'.format(As.shape,bs.shape))
    xs,offsets = optimal_tikhonov_regularization(As,bs,allow_offset = allow_offset,std_noise = std_noise)
    return xs

####################
##   exp ramp     ##
class Ramp(abc.ABC):
    @abc.abstractmethod
    def eval(self,x):
        pass
    def __call__(self,x):
        return self.eval(x)
    
class ExponentialRamp(Ramp):
    def __init__(self,start,stop,exponent,stop_argument=1):
        self.start = start
        self.stop = stop
        self.stop_argument = stop_argument
        if (stop < start):
            exponent *= exponent/abs(exponent)*-1
        else:
            exponent *= exponent/abs(exponent)
        self.exponent = exponent
        self.A=0
        self.B=0
        self.set_model_parameters()
    def set_model_parameters(self):
        self.A = (self.start-self.stop)/(1-np.exp(self.exponent*self.stop_argument))
        self.B = self.start-self.A
    def eval(self,x):
        if self.start>self.stop:
            val = np.maximum(self.A*np.exp(x*self.exponent)+self.B,self.stop)
        else:
            val = np.minimum(self.A*np.exp(x*self.exponent)+self.B,self.stop)
        return val

class LinearRamp(Ramp):
    def __init__(self,start,stop=False,slope=False,default_start=False,default_stop =False):
        self.default_stop = default_stop
        self.default_start = default_start
        #log.info(f'input = {[start,stop,slope]}')
        if not isinstance(start,(list,tuple)):
            self.start = (start,0)
        else:
            self.start = start
            
        self.undefined = False
        if not np.issubdtype(np.array(self.start[0]).dtype,np.number):
            if default_start == False:
                self.undefined = True
            else:
                self.start= (default_start,0)
            
        self.stop,self.stop_is_defined = self.parse_stop(stop)
        self.slope_is_defined = not isinstance(slope,bool)
        self.slope = slope
        #log.info(f'lin ramp options = {[self.start,self.stop,self.slope]}')
        #assert self.stop_is_defined or self.slope_is_defined,'Linear ramp requires either a stop value or a slope value given on instanciation.'
        if not self.undefined:
            self.set_model_parameters()

    def parse_stop(self,stop):
        stop_is_valid = False
        if isinstance(stop,(list,tuple)):
            stop_value_is_number = np.issubdtype(np.array(stop[0]).dtype,np.number)
            if not stop_value_is_number:
                default_is_number = np.issubdtype(np.array(self.default_stop).dtype,np.number)
                if default_is_number:
                    stop[0] = self.default_stop
                    stop_value_is_number = True
            stop_argument_is_number =  np.issubdtype(np.array(stop[1]).dtype,np.number)
            if stop_argument_is_number and stop_value_is_number:
                stop_after_start = stop[1]>=self.start[1]
                if stop_after_start:
                    stop_is_valid = True
        if not stop_is_valid:
            stop = False
        return stop,stop_is_valid
    def set_model_parameters(self):
        start,stop,slope = self.start,self.stop,self.slope
        if (not self.stop_is_defined) and (not self.slope_is_defined):
            self.A=0
            self.B=start[0]
        elif self.stop_is_defined:
            self.C=stop[0]
            if (stop[1]-start[1])==0:
                self.A=0
            else:
                self.A=(stop[0]-start[0])/(stop[1]-start[1])
            if self.slope_is_defined:
                self.A = slope
        elif slope == 0:
            self.C = np.nan
            self.A = slope
        else:
            self.C=np.sign(slope)*np.inf
            self.A = slope
        self.B = start[0]-self.A*start[1]
                
    def eval(self,x):
        if not self.undefined:
            val = self.A*x+self.B
            if self.A<0:
                val = max(val,self.C)
            elif self.A>0:
                val = min(val,self.C)
        else:
            val = np.nan
        return val
#############
## linalg  ##
def distance_from_line_2d(line_points,grid):
    p1,p2 = line_points
    dist = p2-p1
    dist_rot = np.array([[0,1],[-1,0]])@dist
    grid = grid-p1
    dist=np.sum(grid*dist_rot[None,None,:],axis=-1)
    return dist
    

############
###statistics###
def relMeanSquareError(expected,approximation):
    expected=np.asarray(expected)
    approximation=np.asarray(approximation)

    absExpected=np.abs(expected)
    divisors=np.where(absExpected<1e-15,1,np.square(np.abs(expected)))
    differences=np.where(absExpected<1e-15,0,np.square(np.abs(expected-approximation)))

    errors=(differences/divisors).flatten()
    meanError=np.sqrt(1/len(differences)*np.sum(differences/divisors))
    return meanError
    

###############################################
## Polar/Spherical FFT reciprocity relations ##
def polar_spherical_dft_reciprocity_relation_radial_cutoffs_old(cutoff:float,n_points:int,pi_in_q = True):
    '''
    Reciprocity relation between the real and reciprocal cutoff.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.    
    $2*R*2*Q = 2*N$
    '''
    if pi_in_q:
        other_cutoff = np.pi*n_points/cutoff
    else:
        other_cutoff = n_points/(2*cutoff)
    return other_cutoff

def polar_spherical_dft_reciprocity_relation_radial_cutoffs(cutoff:float,n_points:int,reciprocity_coefficient=np.pi):
    '''
    Reciprocity relation between the real and reciprocal cutoff.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.    
    $2*R*2*Q = 2*N$
    '''
    other_cutoff = reciprocity_coefficient*n_points/cutoff    
    return other_cutoff

def _pi_in_q__to__reciprocity_coefficient(pi_in_q):
    if pi_in_q:
        return np.pi
    else:
        return 1/2

    
def polar_spherical_dft_reciprocity_relation_radial_steps(step:float,n_points:int,pi_in_q = True):
    '''
    Reciprocity relation between the real and reciprocal step sizes.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.
    $\delta R*\delta Q = 1/(2*N)$    
    '''
    if pi_in_q:
        other_step = np.pi*1/(n_points*step)
    else:
        other_step = 1/(2*n_points*step)
    return other_step
    
def polar_spherical_dft_reciprocity_relation_radial_step_cutoff(step:float,n_points:int,pi_in_q = True):
    '''
    Reciprocity relation between the real and reciprocal step sizes.
    If pi_in_q is true it is assumed that the reciprocal unit is 2Pi/length otherwise it is just 1/length.
    $\delta R*\delta Q = 1/(2*N)$    
    '''
    if pi_in_q:
        other_cutoff = np.pi/step
    else:
        other_cutoff = 1/(2*step)
    return other_cutoff
    

#assumes sperical coordinate grid where r,phi are uniformely sampled and cos(theta) are gauss legendre nodes.
# grid coords are (r,theta,phi)
class SphericalIntegrator():
    def __init__(self,grid):
        self.n_r,self.n_theta,self.n_phi = grid.shape[:-1]
        self.grid = grid
        self.max_r = np.max(grid[:,0,0,0])
        self.norm = 4/3*np.pi*self.max_r**3
        self.gauss_weights = roots_legendre(self.n_theta)[1]
        self.integrate,self.integrate_normed = self.generate_integration_routines()
        
        

    def generate_integration_routines(self):
        pi = np.pi
        w = self.gauss_weights
        rs = self.grid[:,0,0,0]
        n = self.n_theta
        norm = self.norm
        def integrate(values):
            w_shape = (1,) + w.shape + (1,)*(values.ndim - 3)
            rs_shape = rs.shape + (1,)*(values.ndim - 3)
            s2_int = pi/n*np.sum(w.reshape(w_shape)*np.sum(values,axis=2),axis = 1)
            r_int = np.trapz(s2_int * (rs**2).reshape(rs_shape) , x = rs,axis = 0)
            return r_int
        def integrate_normed(values):
            return integrate(values)/norm
        return integrate,integrate_normed
    def L2_norm(self,values):
        return self.integrate(values*values.conj())
                

class PolarIntegrator():
    '''
    Assumes polar coordinate grid where r,phi are uniformely sampled.
    Grid coords are (r,phi)
    '''
    def __init__(self,grid):
        self.n_r,self.n_phi = grid.shape[:-1]
        self.grid = grid
        self.max_r = np.max(grid[:,0,0])
        self.norm = np.pi*self.max_r**2
        self.integrate,self.integrate_normed = self.generate_integration_routines()

    def generate_integration_routines(self):
        rs = self.grid[:,0,0]
        norm = self.norm
        phis = self.grid[0,:,1]
        def integrate(values):
            rs_shape = rs.shape + (1,)*(values.ndim - 2)
            s_int = np.trapz(values , x = phis,axis = 1)
            r_int = np.trapz(s_int * rs.reshape(rs_shape) , x = rs,axis = 0)
            return r_int
        def integrate_normed(values):
            return integrate(values)/norm
        return integrate,integrate_normed
    def L2_norm(self,values):
        return self.integrate(values*values.conj())


class RadialIntegrator():
    def __init__(self,radial_points,dimension):        
        self.n_r = len(radial_points)
        self.radial_points = radial_points
        self.dimension = dimension
        self.max_r,self.min_r = radial_points.max(),radial_points.min()
        self.norm = np.pi*self.max_r**dimension-np.pi*self.min_r**dimension
        self.rweights = self.radial_points**(self.dimension-1)
        self.integrate,self.integrate_normed = self.generate_integration_routines()
    def generate_integration_routines(self):
        rs = self.radial_points
        w = self.rweights
        norm = self.norm
        def integrate(values,axis = -1):
            #rs_shape = rs.shape + (1,)*(values.ndim - 2)
            len_shape = len(values.shape)
            weight_slice=(None,)*(axis%len_shape) + (slice(None),) + (None,)*(len_shape-axis%len_shape-1)
            r_int = np.trapz(values * w[weight_slice] , x = rs,axis = -1)
            return r_int
        def integrate_normed(values):
            return integrate(values)/norm
        return integrate,integrate_normed
    
    def L2_norm(self,values,axis=-1):
        return self.integrate(values*values.conj(),axis = axis)        
    
class Cacheaware_numpy:
    @staticmethod
    def conjugate(x,*args,block_size=None,out=None,**kwargs):
        if out == None:
            out = np.empty(x)
            

################
###statistics###

def relMeanSquareError(expected,approximation):
    expected=np.asarray(expected)
    approximation=np.asarray(approximation)

    absExpected=np.abs(expected)
    divisors=np.where(absExpected<1e-15,1,np.square(np.abs(expected)))
    differences=np.where(absExpected<1e-15,0,np.square(np.abs(expected-approximation)))

    errors=(differences/divisors).flatten()
    meanError=np.sqrt(1/len(differences)*np.sum(differences/divisors))
    return meanError


def nan_mask_decorator(func):
    def masked_func(*args,fill_value = 0.0, return_mask = False, **kwargs):
        with warnings.catch_warnings():
            warnings.filterwarnings('ignore')
            result = np.atleast_1d(func(*args,**kwargs))
        nan_mask = np.isnan(result)
        result[nan_mask] = fill_value
        if return_mask:
            return result, ~nan_mask
        else:
            return result
    return masked_func

#wrapper for numpy mean and std
@nan_mask_decorator
def masked_mean_old(*args,**kwargs):
    return np.mean(*args,**kwargs)

@nan_mask_decorator
def masked_std(*args,**kwargs):
    return np.std(*args,**kwargs)
    
def combine_means(means,n_data_points):
    n_total_data_points = np.sum(n_data_points)    
    combined_mean = np.sum(tuple(mean*n_data_points for mean,n_data_points in zip(means,n_data_points)))/n_total_data_points
    return combined_mean,n_total_data_points

def masked_mean(data,mask,axis = 0):
    counts = np.sum(mask,axis = axis)
    counts_non_zero = (counts!=0)
    data_sum = np.sum(data,where=mask,axis = axis)
    data_sum[counts_non_zero]=data_sum[counts_non_zero]/counts[counts_non_zero]
    return data_sum,counts
def combine_means_2D(means,counts,axis = 0):
    means = np.asarray(means)
    counts = np.asarray(counts)
    #log.info('means shape in combine ={}'.format(means.shape))
    total_counts = np.sum(counts,axis = axis )
    non_zero_counts = total_counts!=0
    combined_mean = np.sum(means*counts,axis = axis)
    combined_mean[non_zero_counts] /= total_counts[non_zero_counts]
    return combined_mean,total_counts

def masked_variance(data,mask,axis = 0):
    var = np.var(data,where=mask,axis=axis)
    counts = np.sum(mask,axis = axis)
    nan_mask = np.isnan(var)
    var[nan_mask]=0
    return var,counts

def combine_variances_ND(_vars,means,counts,axis=0):
    _vars = np.moveaxis(_vars,axis,0)
    means = np.moveaxis(means,axis,0)
    counts = np.moveaxis(counts,axis,0)
    combined_variance = _vars[0].copy()
    combined_mean = means[0].copy()
    combined_counts = counts[0].copy()
    for var,mean,c in zip(_vars[1:],means[1:],counts[1:]):
        next_combined_mean,denom = combine_means_2D([combined_mean,mean],[combined_counts,c])
        non_zero = denom !=0 
        N1 = combined_counts[non_zero]/denom[non_zero]
        N2 = c[non_zero]/denom[non_zero]
        N3 = (combined_counts*c)[non_zero]/(denom**2)[non_zero]

        a = N1*combined_variance[non_zero]
        combined_variance[non_zero]=a+N2*var[non_zero]+N3*np.square(combined_mean-mean)[non_zero]
        combined_mean = next_combined_mean
        combined_counts = denom
    return combined_variance,combined_mean,combined_counts


### connected area ###
#Simple algorithm to find te area of equal values in a 2d array/image
def _get_connected_area_periodic(image,point,step,value,shape,connected_points,visited_mask):
    #print(point)
    #print(im[point])
    if visited_mask[point]:
        return connected_points
    if image[point]==value:
        connected_points+=[point]
        visited_mask[point]=True
        
        orth_step = ((step[0]+1)%2,(step[1]+1)%2)
        next_steps = (step,orth_step,(-orth_step[0],-orth_step[1]))
        next_points = (((point[0]+s[0])%shape[0],(point[1]+s[1])%shape[1]) for s in next_steps)
        for p,s in zip(next_points,next_steps):
            _get_connected_area_periodic(image,p,s,value,shape,connected_points,visited_mask)
    return connected_points

def _get_connected_area(image,point,step,value,shape,connected_points,visited_mask):
    #print(point)
    #print(im[point])   
    if visited_mask[point]:
        return connected_points

    if image[point]==value:
        connected_points+=[point]
        visited_mask[point]=True
        
        orth_step = ((step[0]+1)%2,(step[1]+1)%2)
        next_steps = (step,orth_step,(-orth_step[0],-orth_step[1]))
        next_points = ((point[0]+s[0],point[1]+s[1]) for s in next_steps)
        for p,s in zip(next_points,next_steps):
            if (p[0]>=0 and p[0]<shape[0] and p[1]>=0 and p[1]<shape[1]):
                _get_connected_area(image,p,s,value,shape,connected_points,visited_mask)
    return connected_points

def find_connected_component(image,start,periodic = False,return_mask = False):
    start=tuple(start)
    value = image[start]
    shape = image.shape
    next_steps= np.array(((1,0),(-1,0),(0,1),(0,-1)))
    
    if periodic:
        find_connected_points = _get_connected_area_periodic
        next_points = (next_steps+np.array(start)[None,:])%np.array(shape)[None,:]
    else:
        find_connected_points = _get_connected_area
        next_points = next_steps+np.array(start)[None,:]
        negative_mask = np.sum(next_points<0,axis = -1).astype(bool)
        to_high_mask = np.sum(next_points>=np.array(shape)[None,:],axis = -1).astype(bool)
        point_mask = ~(negative_mask | to_high_mask)
        next_points = next_points[point_mask]
        next_steps = next_steps[point_mask]
    
    
    visited_mask = np.zeros(shape,dtype = bool)
    connected_points=[tuple(start)]
    for point, step in zip(next_points,next_steps):
        find_connected_points(image,tuple(point),step,value,shape,connected_points,visited_mask)
    if return_mask:
        mask = np.zeros_like(visited_mask)
        mask[tuple(np.stack(connected_points,axis = 1))]=True
        return connected_points,mask
    else:
        return connected_points

def get_test_function(support=[-1,1],slope=1):
    center = np.mean(support)
    size = support[1]-center
    #print('size = {}'.format(size))
    #print('center = {}'.format(center))
    def test_function(data):
        non_zero = ((data>support[0]) & (data<support[1]))
        values = np.zeros_like(data)
        values[non_zero]=np.exp(-slope*size**2/(size**2 - (data[non_zero]-center)**2))
        return values
    return test_function

def get_test_function2(support=[-1,-0.5,0.5,1],slope=1):
    assert support[0]<upport[1]<upport[2]<upport[3]
    #print('size = {}'.format(size))
    #print('center = {}'.format(center))
    def f(x):
        return np.exp(-slope*(1/(1-(x-1)**2)-1))
    def test_function(data):
        non_zero = ((data>support[0]) & (data<support[-1]))
        values = np.zeros_like(data)
        x=data[non_zero]
        values[non_zero]=f((x-support[0])/(support[1]-support[0]))*f((x-support[3])/(support[3]-support[2]))
        return values
    return test_function

#### solve procrustes problem ###

def solve_procrustes_problem(V1,V2):
    '''
    Finds the unitary matrix U that minimizes ||V1-V2U||.
    by computing the svd of V2^\dagger V1
    '''
    #log.info(f'trasposed shapes = {V2.T.shape} @ {V1.shape}')
    return np.matmul(*np.linalg.svd((V2.conj().T) @ V1,full_matrices = False)[::2])

def midpoint_rule(samples,uniform_sampling_points,**kwargs):
    step = uniform_sampling_points[1]-uniform_sampling_points[0]
    #N = len(uniform_sampling_points)
    integral = step*np.sum(samples,**kwargs)
    return integral


def psd_back_substitution(cn,ppm):
    bl = np.zeros(cn.shape,dtype = complex)
    cn = cn.copy()
    pm = ppm.copy()
    for i in np.arange(0,cn.shape[-1])[::-1]:
        bl[...,i] = nearest_positive_semidefinite_matrix(cn[...,-1]/pm[...,-1,-1])
        cn = (cn[...,:-1]-bl[...,i,None]*pm[...,:-1,-1])
        pm = pm[...,:-1,:-1]
    return bl

def back_substitution(cn,ppm):
    bl = np.zeros(cn.shape,dtype = complex)
    cn = cn.copy()
    pm = ppm.copy()
    for i in np.arange(0,cn.shape[-1])[::-1]:
        bl[...,i] = cn[...,-1]/pm[...,-1,-1]
        cn = (cn[...,:-1]-bl[...,i,None]*pm[...,:-1,-1])
        pm = pm[...,:-1,:-1]
    return bl



from xframe.presenters import matplotlibPresenter as mpres
import numpy as np
from xframe.library.mathLibrary import spherical_to_cartesian



def Intensity_histogramm(values,n_patterns,run,caption):
    layout = {
        'title': 'Run:{} Intensity Histogramm \n {} considered patterns'.format(run,n_patterns),
        'x_label': 'Mean Intensity',
        'y_label': 'Frequency',
        'caption': caption
    }
    return mpres.hist1D.get_fig(values,y_scale = 'log',layout = layout)

    
def mean_2d(mean,run,caption,geometry,lab_grid=False):
    if lab_grid:
        print_grid = geometry['framed_lab_pixel_grid']
    else:
        print_grid = spherical_to_cartesian(geometry['framed_pixel_grid'])
    print_mask = geometry['framed_mask']
    layout = {
        'title': 'Run:{} Pixelwise Mean'.format(run),
        'x_label': 'x',
        'y_label': 'y',
        'caption': caption
    }
    return mpres.agipd_heat.get_fig(np.abs(mean),print_grid,print_mask,scale = 'log',layout = layout,vmin=1)

def max_2d(maximum,run,caption,geometry,lab_grid=False):
    if lab_grid:
        print_grid = geometry['framed_lab_pixel_grid']
    else:
        print_grid = spherical_to_cartesian(geometry['framed_pixel_grid'])
    print_mask = geometry['framed_mask']
    layout = {
        'title': 'Run:{} Pixelwise Maximum'.format(run),
        'x_label': 'x',
        'y_label': 'y',
        'caption': caption
    }
    return mpres.agipd_heat.get_fig(maximum,print_grid,print_mask,scale = 'log',layout = layout,vmin = 1)

def std_2d(maximum,run,caption,geometry,lab_grid=False):
    if lab_grid:
        print_grid = geometry['framed_lab_pixel_grid']
    else:
        print_grid = spherical_to_cartesian(geometry['framed_pixel_grid'])
    print_mask = geometry['framed_mask']
    layout = {
        'title': 'Run:{} Pixelwise Standard Deviation'.format(run),
        'x_label': 'x',
        'y_label': 'y',
        'caption': caption
    }
    return mpres.agipd_heat.get_fig(maximum,print_grid,print_mask,scale = 'log',layout = layout,vmin=0.6)



